;; This serves as a wrapper around games.sc

;; You can either compile this or load it into qsci
;; Typical commands are:
;;   (set! *game* 'hexapawn)
;;   (set! *output-type* 'outcome)
;;   (set! *output-type* 'legal-move)
;;   (set! *variant* 'normal)
;;   (setup-new-game '(3))
;;   (generate-game '(3))

(module main (with QobiScheme) (main main))

(eval-when (eval) (load "games.sc"))
(include "games.sc")

(define (pick-up square)
 (when *holding* (message "You are already holding something") (abort))
 (set! *holding* (square-contents square))
 (set-square-contents! square #f)
 (redraw-display-pane))

(define (put-down square)
 (unless *holding* (message "You are not holding anything") (abort))
 (set-square-contents! square *holding*)
 (set! *holding* #f)
 (redraw-display-pane))

(define (clean-up)
 (let ((froms (remove-if-not (lambda (square)
			      (and (square-contents square)
				   (eq? (first (square-name square)) 'board)))
			     *squares*))
       (tos (remove-if-not (lambda (square)
			    (and (not (square-contents square))
				 (eq? (first (square-name square)) 'staging)))
			   *squares*)))
  (for-each (lambda (from to)
	     (pick-up from)
	     (put-down to))
	    froms
	    tos)))

(define *square* #f)
(define *holding* #f)


(eval-when (eval)
  (set! abort
	(lambda ()
	  (*abort-continuation* #f)))
  (define (getpid) 1)
  (define (getenv x) "x"))

(define-c-external (c-srand int) int "srand")

(define *nr-games* #f)
(define *game-size* #f)
(define *noise* #f)
(define *seed* #f)

(define (main args)
  (set! abort
	(lambda ()
	  (*abort-continuation* #f)))
  (if (= (length args) 8)
      (begin
	(set! *game* (string->symbol (string-upcase (list-ref args 1))))
	(set! *variant*
	      (string->symbol (string-upcase (list-ref args 2))))
	(set! *output-type*
	      (string->symbol (string-upcase (list-ref args 3))))
	(set! *nr-games* (string->number (list-ref args 4)))
	(set! *game-size* (string->number (list-ref args 5)))
	(set! *noise* (string->number (list-ref args 6)))
	(set! *seed* (string->number (list-ref args 7)))
	(c-srand *seed*)
	(if (and *noise* (zero? *noise*))
	    (display
	     (prolog-program->string
	      (generate-prolog-games *nr-games* `(,*game-size*))))
	    (display
	     (prolog-program->string
	      (generate-noisy-prolog-games *nr-games* `(,*game-size*) *noise*))))
	(newline))
      (begin
	(display "games-generator <game-type> <game-version> <output-type (legal-move|outcome)> <nr-games> <board size> <integer errors per game> <rng seed>")(newline)
	(display "(game-type variants..) supported are:")(newline)
	(map (lambda (x)
	       (display (string-downcase (format #f "~a~%" x))))
	     *games-supported*)
	(newline))))
