(include "QobiScheme.sch")
(include "c-externals.sc")

(define-structure square size x y contents name)

(define *variant* 'normal)
(define *output-type* 'legal-move)
(define (create-square x y contents name) (make-square 40 x y contents name))
(define *squares* '())
(define *game* #f)
(define *games-supported* '((tic-tac-toe normal) (hexapawn normal variant-a variant-b variant-c variant-d)))
(define *current-game* '())
(define *current-game-settings* #f)
(define *percent-random-play* 100)
(define *playing-area-size* #f)


;;; Randomization

(define (flip-piece-3 x e y a)
  (cond ((eq? a x) (random-member `(,e ,y)))
	((eq? a e) (random-member `(,x ,y)))
	((eq? a y) (random-member `(,e ,x)))))

(define (flip-board-piece pieces a)
  (flip-piece-3
   (first pieces)
   (second pieces)
   (third pieces)
   a))

(define (game->pieces)
 (case *game*
  ((tic-tac-toe) '(x #f o))
  ((hexapawn) '(w #f b))
  (else (fuck-up))))

(define (drop n l) (if (zero? n) l (drop (- n 1) (cdr l))))

(define (find-nth-if n p l)
  (if (zero? n)
      (find-if p l)
      (find-nth-if (- n 1) p (drop (+ (position-if p l) 1) l))))

(define (random-integers max nr)
  (let loop ((numbers '()))
    (if (= (length numbers) nr)
	numbers
	(loop (union `(,(random-integer max)) numbers)))))

(define (game-size g) (count-if board? g))

(define (add-game-noise! nr-points game pieces)
  (let* ((nr-boards (count-if board? game))
	 (nr-positions (* nr-boards 9))
	 (locations (random-integers nr-positions nr-points)))
    (map
     (lambda (x)
       ((lambda (y)
	  (display (vector-ref y 4))(newline)
	  (display x)(newline)
	  (vector-set!
	   y
	   4
	   (flip-board-piece (game->pieces) (vector-ref y 4))))
	(list-ref
	 (board-playing-area
	  (find-nth-if (inexact->exact (/ x 9))
		       board?
		       game))
	 (+ 6 (modulo x 9)))))
     locations)
    game))


;;; Prolog

;;; clauses are represented as
;;; (name (head-vars) (call body))
;;; a rather literal translation of the scheme equivalent
;;; (clause->prolog-string '("tenu" ((list "D") () "N") (("digit" "D" "N"))))
;;; "tenu([D],[],N) :- digit(D,N)."

(define (caddadr x) (car (cddadr x)))

(define (symbol->lowercase-string s)
  (string-downcase (symbol->string s)))

(define (prolog-clause->string c)
  (if (and (string? (car c)) (null? (cdr c)))
      (car c)
      (string-append
       (if (null? (car c))
	   ""
	   (string-append
	    (prolog-clause-body->string (car c));)
	    (if (null? (cadr c))
		""
		(string-append
		 "("
		 (string-concat "," (map prolog-clause-body->string (cadr c)))
		 ")"))))
       (if (or (null? (cddr c))
	       (null? (caddr c)))
	   ""
       (string-append
	" :- "
	(string-concat "," (map prolog-clause-body->string (caddr c)))))
       (if (equal? (last c) '?)
	   "?"
	   "."))))

(define (string-concat delim ss)
  (reduce (lambda (x y) (string-append x delim y)) ss ""))

(define (prolog-clause-body->string c)
  (cond ((null? c) "[]")
	((symbol? c) (symbol->string c))
	((string? c) c)
	((number? c) (number->string c))
	((or (equal? (car c) "*")
	     (equal? (car c) "+")
	     (equal? (car c) "-")
	     (equal? (car c) "/"))
	 (string-append
	  (prolog-clause-body->string (cadr c))
	  (car c)
	  (prolog-clause-body->string (caddr c))))
	((eq? (car c) 'list-pat)
	 (unless (> (length (cdr c)) 1) (panic "bad prolog atom ~a" c))
	 (string-append
	  "["
	  (let loop ((l (cddr c)) (s (prolog-clause-body->string (cadr c))))
	     (if (null? (cdr l))
		 (string-append
		  s
		  (string-append "|" (prolog-clause-body->string (car l))))
		 (loop (cdr l)
		       (string-append s "," (prolog-clause-body->string (car l))))))
	  "]"))
	((eq? (car c) 'list)
	 (string-append
	  "["
	  (string-concat
	   ","
	   (let loop ((l (cdr c)))
	     (if (null? l)
		 '()
		 (cons (prolog-clause-body->string (car l)) (loop (cdr l))))))
	  "]"))
	(else
	 (string-append
	  (prolog-clause-body->string (car c))
	  "("
	  (string-concat
	   ","
	   (map prolog-clause-body->string (cdr c)))
	  ")"))))

(define prolog->string prolog-clause->string)
(define (prolog-program->string p) (string-concat (list->string '(#\newline)) (map prolog->string p)))

(define (modeh b) `(() () (("modeh" ,@b)) ?))
(define (modeb b) `(() () (("modeb" ,@b)) ?))

(define (prolog-input x) (string-append "+" x))
(define (prolog-output x) (string-append "-" x))
(define (prolog-constant x) (string-append "#" x))

(define (prolog-set b) `(() () (("set" ,@b)) ?))

(define (positive-only-learning) (prolog-set '("posonly")))

(define (unlines lines)
  (reduce (lambda (x y) (string-append x (format "~%") y)) lines ""))

(define (char-downcase? c) (eq? (char-downcase c) c))

(define (prolog-clause-starts? line)
  (let ((unline (string->list line)))
    (and (not (null? unline))
	 (char-downcase? (car unline))
	 (char-alphabetic? (car unline)))))

(define (skip-until-prolog-clause lines_)
  (let loop ((lines lines_))
    (if (or
	 (null? lines)
	 (prolog-clause-starts? (car lines)))
	lines
	(loop (cdr lines)))))

(define (prolog-clause-ends? line)
  (findq #\. (string->list line)))

(define (prolog-join-clauses lines)
  (if (and (not (null? lines))
	   (prolog-clause-starts? (car lines))
	   (not (prolog-clause-ends? (car lines))))
      (prolog-join-clauses
       (cons (string-append (car lines)
			    (cadr lines))
	     (cddr lines)))
      lines))

(define (split-at x l)
  (let ((loc (position x l)))
    (if (or (null? l) (not loc))
	`(,l)
	(cons (sublist l 0 loc)
	      (split-at x (sublist l (+ loc 1) (length l)))))))

(define (split-string-at x s)
  (map list->string  (split-at x (string->list s))))

(define (split-clause line) (map (lambda (x)
				   (if (prefix? "-" x)
				       (substring x 1 (string-length x))
				       x))
				 (split-string-at #\: line)))

(define (string-position x s)
  (position x (string->list s)))

(define (prolog-read-end-marker line)
  `(,(substring line 0 (string-position #\. line)) "."))

(define (prolog-read-identifier lst)
  (let loop ((id '()) (l lst))
    (cond ((null? l)
	   `(,(list->string (reverse id)) ,l))
	  ((eq? #\( (car l))
	   (let ((x (prolog-read-args l)))
	     `((,(list->string (reverse id)) ,@(car x)) ,(cadr x))))
	  ((not (prolog-id-char? (car l)))
	   `(,(list->string (reverse id)) ,l))
	  (else
	   (loop (cons (car l) id) (cdr l))))))

(define (prolog-id-char? x)
  (or (char-alphanumeric? x)
      (eq? #\_ x)))

(define (prolog-read-args lst)
  (let loop ((ids '()) (l (if (eq? #\( (car lst))
			      (cdr lst)
			      lst)))
    (cond ((null? l)
	   `(,(reverse ids) ,l))
	  ((eq? #\) (car l))
	   `(,(reverse ids) ,(cdr l)))
	  ((eq? #\, (car l))
	   (loop ids (cdr l)))
	  ((char-whitespace? (car l))
	   (loop ids (cdr l)))
	  ((prolog-id-char? (car l))
	   (let ((x (prolog-read-identifier l)))
	     (loop (cons (car x) ids)  (cadr x))))
	  (else `(,(reverse ids) ,l)))))

(define (prolog-read-clause line)
  (let ((split (split-clause line)))
    (unless (= (length split) 2) #f)
    `(,@((lambda (x)
	   `(,(car x)
	     ,(cdr x)))
	 (car (prolog-read-identifier (string->list (car split)))))
      ,(car (prolog-read-args (string->list (cadr split)))))))

(define (read-prolog p)
  (let ((c (prolog-join-clauses (skip-until-prolog-clause p))))
    (if (null? c)
	'()
	(cons (prolog-read-clause (car c))
	      (read-prolog (cdr c))))))


;;; General interface
;;; generate-initial :: [parameters] -> (initial, start-player)
;;; generate-full-game :: initial -> start-player -> generator -> game
;;; generate-prolog :: game -> prolog-program
;;; So we generate a list of size two
;;; a representation of the actions for positive-only learning
;;; a representation of the outcomes for normal learning
;;; todo the ->prolog actions should be implemented

(define-structure board player playing-area)
(define (player-update board f)
  (make-board (f board) (board-playing-area board)))

(define (transpose-square s)
  (let* ((ss (copy-vector s))
	 (x (square-x ss))
	 (y (square-y ss)))
    (set-square-x! ss x)
    (set-square-y! ss y)
    ss))

(define (square-name-update f s)
  (let ((ns (copy-vector s)))
    (set-square-contents! ns (f (square-contents s)))
    ns))

(define (copy-squares squares)
  (map (lambda (vector) (copy-vector vector)) squares))

(define (copy-vector vector) (list->vector (vector->list vector)))

(define (add-board! player area)
  (set! *current-game*
	(append *current-game*
		`(,(make-board player (copy-squares area))))))
(define (add-move! player move)
  (set! *current-game* (append *current-game* `((move ,player ,move)))))

(define (*squares*->board player)
  (make-board player (*squares*->playing-area)))

(define (squares->board squares player)
  (make-board player (squares->playing-area squares)))

(define (squares->playing-area squares)
  (map-n (lambda (i)
	   (map-n (lambda (j)
		    ((case *game*
		       ((tic-tac-toe) tic-tac-toe-get-square-contents)
		       ((hexapawn) hexapawn-get-square-contents)) i j squares))
		  *playing-area-size*))
	 *playing-area-size*))

(define (*squares*->playing-area) (squares->playing-area *squares*))

;; Game independent

(define (moves board)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-moves board))
  ((hexapawn) (hexapawn-moves board))
  (else (fuck-up))))

(define (make-move move board)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-make-move move board))
  ((hexapawn) (hexapawn-make-move move board))
  (else (fuck-up))))

(define (win0 board)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-win0 board))
  ((hexapawn) (hexapawn-win0 board))
  (else (fuck-up))))

(define (move->actions player move)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-move->actions player move))
  ((hexapawn) (hexapawn-move->actions player move))
  (else (fuck-up))))

(define (move->prolog move prev next)
 (case *game*
  ((tic-tac-toe) (tic-tac-toe-move->prolog move prev next))
  ((hexapawn) (hexapawn-move->prolog move prev next))
  (else (fuck-up))))

(define (maximize f l limit)
 (define (loop best-so-far l)
  (cond ((>= best-so-far limit) 1)
	((null? l) best-so-far)
	(else (loop (max (f (first l) best-so-far) best-so-far) (rest l)))))
 (loop -1 l))

(define (win*-alpha/beta board limit)
  (let ((player-board (board-player board))
	(win0-board (win0 board)))
    (cond ((not (zero? win0-board)) win0-board)
	  ((null? (moves board)) 0)
	  (else
	   (* player-board
	      (maximize
	       (lambda (move limit)
		 (* player-board
		    (win*-alpha/beta
		     (make-move move board) (* player-board limit))))
	       (moves board)
	       (* player-board limit)))))))

(define (win* board) (win*-alpha/beta board (board-player board)))

(define (optimal-moves board)
  (let ((board-winner (win* board))
	(player (board-player board)))
    (remove-if-not (lambda (move)
		     (= (* player (win* (make-move move board)))
			(* player board-winner)))
		   (moves board))))

(define (draw? board) (null? (moves board)))

(define (play! player)
 (let ((board (*squares*->board player)))
  (when (or (not (zero? (win0 board))) (null? (moves board)))
    (abort))
  (for-each (lambda (action) (action))
	    (begin
	      (let ((move (if (> *percent-random-play* (random-integer 99))
			      (random-member (moves board))
			      (random-member (optimal-moves board)))))
		(add-move! player move)
		(move->actions player move))))
  (add-board! (* -1 player) *squares*)))

(define (game->prolog game)
  (case *game*
    ((tic-tac-toe) (tic-tac-toe-game->prolog game))
    ((hexapawn) (hexapawn-game->prolog game))))

(define (board->prolog board name)
  (case *game*
    ((tic-tac-toe) (tic-tac-toe-board->prolog board name))
    ((hexapawn) (hexapawn-board->prolog board name))))

(define (simple-prolog-negative-outcomes board name)
  (append
   (map
    (lambda (player)
      `(() () ("outcome" (win ,player))))
    (game-board-players board))
   `((() () ("outcome" draw)))))

(define (cadaddr x) (cadr (caddr x)))
(define (simple-prolog->outcome outcome) (cadaddr outcome))

(define (simple-fixup-board board)
  (make-board
   (board-player board)
   (squares->playing-area (board-playing-area board))))

(define (simple-outcome->prolog board)
  (let* ((board (simple-fixup-board board))
	 (player (board-player board))
	 (board-area (board-playing-area board))
	 (outcome (win0 board))
	 (moves (moves board)))
    (append
     (if (= 1 outcome)
	 `(("outcome" (,(simple-player->prolog player)
		       ,(simple-board->prolog board-area)
		       "x_wins")))
	     `((() () (("outcome"
			,(simple-player->prolog player)
			,(simple-board->prolog board-area)
			"x_wins")))))
;	     '()))
     (if (= -1 outcome)
	 `(("outcome" (,(simple-player->prolog player)
		       ,(simple-board->prolog board-area)
		       "o_wins")))
	     `((() () (("outcome"
			,(simple-player->prolog player)
			,(simple-board->prolog board-area)
			"o_wins"))))))))

(define (next-player board)
  (case *game*
    ((tic-tac-toe) (tic-tac-toe-next-player board))
    ((hexapawn) (hexapawn-next-player board))
    (else (fuck-up))))

(define (generate-game settings)
  (setup-new-game settings)
  (call-with-current-continuation
   (lambda (c)
     (set! *abort-continuation* c)
     (let loop () (play! 1) (play! -1) (loop))))
  *current-game*)

(define (generate-board-prolog size)
  (apply append
	 (map-n (lambda (n)
		  `(("row" (,(string-append "r" (number->string n))))
		    ("col" (,(string-append "c" (number->string n))))))
		size)))

(define (restart-game) (setup-new-game *current-game-settings*))

(define (setup-new-game settings)
  (set! *current-game* '())
  (set! *current-game-settings* settings)
  (case *game*
    ((tic-tac-toe) (tic-tac-toe-setup-new-game settings)
     (add-board! (tic-tac-toe-initial-player settings) *squares* ))
    ((hexapawn)
     (hexapawn-setup-new-game settings)
     (add-board! (hexapawn-initial-player settings) *squares*))
    (else (fuck-up))))

(define (board-num->string name num) (string-append name (number->string num)))

;; TIC-TAC-TOE

(define (tic-tac-toe-initial-player s) 1)
 (define (tic-tac-toe-get-square-contents i j squares)
  (case (square-contents
	 (find-if (lambda (square)
		   (and (eq? (first (square-name square)) 'board)
			(= i (second (square-name square)))
			(= j (third (square-name square)))))
		  squares))
   ((x) 1)
   ((o) -1)
   ((#f) 0)
   (else (fuck-up))))

(define (tic-tac-toe-setup-new-game settings)
  (set! *game* 'tic-tac-toe)
  (let ((size (car settings)))
    (unless (= size 3)
	    (message "tic-tac-toe only supports games of size 3")
	    (abort))
    (set! *playing-area-size* 3)
    (set! *squares*
	  (list
	   ;; Left staging area
	   (create-square 20 320 'x '(staging left 0))
	   (create-square 20 360 'x '(staging left 1))
	   (create-square 20 400 'x '(staging left 2))
	   (create-square 20 440 'x '(staging left 3))
	   (create-square 20 480 'x '(staging left 4))
	   ;; 3*3 board
	   (create-square 100 360 #f '(board 0 0))
	   (create-square 100 400 #f '(board 1 0))
	   (create-square 100 440 #f '(board 2 0))
	   (create-square 140 360 #f '(board 0 1))
	   (create-square 140 400 #f '(board 1 1))
	   (create-square 140 440 #f '(board 2 1))
	   (create-square 180 360 #f '(board 0 2))
	   (create-square 180 400 #f '(board 1 2))
	   (create-square 180 440 #f '(board 2 2))
	   ;; Right staging area
	   (create-square 260 320 'o '(staging right 0))
	   (create-square 260 360 'o '(staging right 1))
	   (create-square 260 400 'o '(staging right 2))
	   (create-square 260 440 'o '(staging right 3))
	   (create-square 260 480 'o '(staging right 4))))))

(define (old-simple-game->prolog game name)
  (let loop ((game game) (prolog '()) (board-num 0))
    (cond
     ((null? game) prolog)
     ((board? (car game))
      (loop
       (cdr game)
       (append
	prolog
	(board->prolog (car game)
		       (board-num->string name board-num)))
       (+ 1 board-num)))
     ((eq? (caar game) 'move)
      (loop
       (cdr game)
       (append
	prolog
	(move->prolog (cdar game)
			 (board-num->string name (- board-num 1))
			 (board-num->string name board-num)))
       board-num))
     (else (fuck-up)))))

(define (move? x) (eq? (car x) 'move))
(define (simple-move-player move) (second move))

(define (simple-game->prolog game)
  (let loop ((game game) (prolog '()))
    (cond
     ((null? game) prolog)
     ((board? (car game))
      (loop (cdr game)
	    (let ((next-move (find-if move? (cdr game)))
		  (next-board (find-if board? (cdr game))))
	      (cond
	       ((eq? *output-type* 'legal-move)
		(if next-board
		    (append prolog
			    (simple-legal-move (simple-move-player next-move)
					       (car game) next-board))
		    prolog))
	       ((eq? *output-type* 'outcome)
		(append prolog
			(simple-outcome->prolog (car game))))
	       (else (panic "Bad output-type"))))))
     (else (loop (cdr game) prolog)))))

(define (list->prolog-clause x) (cons 'list x))

(define (simple-board->prolog x)
  (list->prolog-clause
   (map (lambda (x)
	  (list->prolog-clause
	   (map (lambda (x)
		  (cond ((= x 1) "x")
			((= x -1) "o")
			((= x 0) "none")
			(else (fuck-up))))
		x)))
	x)))

(define (simple-player->prolog player)
  (cond ((= player 1) "player_x")
	((= player -1) "player_o")
	(else (fuck-up))))

(define (simple-legal-move player x y)
  (let ((a (squares->playing-area (board-playing-area x)))
	(b (squares->playing-area (board-playing-area y))))
    `(("legal_move"
      (,(cond ((= player 1) "player_x")
	      ((= player -1) "player_o")
	      (else (fuck-up)))
       ,@(map (lambda (x)
		(simple-board->prolog x)) `(,a ,b)))))))

(define tic-tac-toe-game->prolog simple-game->prolog)

(define (tic-tac-toe-board->prolog board name) (simple-board->prolog board))

(define (simple-playing-positions->prolog playing-area name piecef)
  (apply
   append
   (map-indexed
    (lambda (row row-num)
      (map-indexed (lambda (entry col-num)
		     `("board_p"
		       (,(symbol->lowercase-string (piecef entry))
			,(number->row row-num)
			,(number->column col-num)
			,name)))
		   row))
    playing-area)))

(define (tic-tac-toe-move->prolog move prev next)
  `(("move" (,(symbol->lowercase-string (simple-move->player move))
	     ,(string-append
	       (symbol->lowercase-string (tic-tac-toe-move->piece-type move))
	       "_act")
	     ,(number->row (tic-tac-toe-move->row move))
	     ,(number->column (tic-tac-toe-move->column move))
	     ,prev
	     ,next))))

(define (number->column n) (format "c~a" n))
(define (number->row n) (format "r~a" n))

(define (tic-tac-toe-player->symbol player)
  (case player
    ((1)  'player1)
    ((-1) 'player2)
    (else (fuck-up))))

(define (simple-move->player move)
  (case (first move)
    ((1)  'player1)
    ((-1) 'player2)
    (else (fuck-up))))

(define (tic-tac-toe-move->piece-type move)
  (tic-tac-toe-piece-num->piece-type (first move)))

(define (tic-tac-toe-piece-num->piece-type piece-num)
  (case piece-num
    ((1)  'x)
    ((0)  'none)
    ((-1) 'o)
    (else (fuck-up))))

(define (tic-tac-toe-move->column move) (cadadr move))
(define (tic-tac-toe-move->row move) (caadr move))

(define (game-player->symbol player)
  (case *game*
    ((tic-tac-toe) (tic-tac-toe-player->symbol player))
    ((hexapawn) (hexapawn-player->symbol player))
    (else (fuck-up))))

(define (game-board-players board)
  (case *game*
    ((tic-tac-toe) (tic-tac-toe-game-board-players board))
    ((hexapawn) (hexapawn-game-board-players board))
    (else (fuck-up))))

(define (tic-tac-toe-game-board-players game) '(player1 player2))
(define (hexapawn-game-board-players game) '(player1 player2))

(define (generate-games settings)
  (map (lambda (setting)
	 (generate-game setting)
	 *current-game*) settings))

(define (concat x) (reduce append x '()))

(define (generate-prolog-games n setting)
  (concat (map-n (lambda (x) (game->prolog (generate-game setting))) n)))

(define (generate-noisy-prolog-games n setting noise)
  (let ((games (map-n (lambda (x) (generate-game setting)) n)))
    (add-game-noise! noise (concat games) (game->pieces))
    (concat (map game->prolog games))))

(define (tic-tac-toe-moves board)
 (define (discard-non-empty row)
  (cond ((null? row) '())
	((zero? (third (first row)))
	 (cons (first row) (discard-non-empty (rest row))))
	(else (discard-non-empty (rest row)))))
 (map (lambda (move) (list (first move) (second move)))
      (reduce append
	      (map-indexed
	       (lambda (row i)
		 (discard-non-empty
		  (map-indexed (lambda (element j) (list i j element)) row)))
	       (board-playing-area board))
	      '())))

(define (list-replace-ith list i x)
  (if (= i 0)
      (cons x (rest list))
      (cons (first list) (list-replace-ith (rest list) (- i 1) x))))

(define (tic-tac-toe-make-move move board)
  (make-board
   (tic-tac-toe-next-player board)
   (list-replace-ith
    (board-playing-area board)
    (first move)
    (list-replace-ith
     (list-ref (board-playing-area board) (first move)) (second move) (board-player board)))))

(define (tic-tac-toe-win0 board)
 (define (at board i j) (list-ref (list-ref (board-playing-area board) i) j))
 (define (win? player)
  (let ((n (length (board-playing-area board))))
   (or
    (some-n (lambda (i) (every-n (lambda (j) (= (at board i j) player)) n)) n)
    (some-n (lambda (j) (every-n (lambda (i) (= (at board i j) player)) n)) n)
    (every-n (lambda (i) (= (at board i i) player)) n)
    (every-n (lambda (i) (= (at board i (- (- n 1) i)) player)) n))))
 (cond ((win? 1) 1)
       ((win? -1) -1)
       (else 0)))

(define (tic-tac-toe-move->actions player move)
 (let ((from (find-if (lambda (square)
		       (and (eq? (first (square-name square)) 'staging)
			    (eq? (square-contents square)
				 (if (= player 1) 'x 'o))))
		      *squares*))
       (to (find-if (lambda (square)
		     (and (eq? (first (square-name square)) 'board)
			  (= (second (square-name square)) (first move))
			  (= (third (square-name square)) (second move))))
		    *squares*)))
   (list (lambda () (pick-up from))
	 (lambda () (put-down to)))))

(define (tic-tac-toe-next-player board)
  (player-update
   board
   (lambda (board) (- (board-player board)))))

;; Hexapawn

(define (hexapawn-get-square-contents i j squares)
  (case (square-contents
	 (find-if (lambda (square)
		   (and (eq? (first (square-name square)) 'board)
			(= i (second (square-name square)))
			(= j (third (square-name square)))))
		  squares))
   ((w) 1)
   ((b) -1)
   ((#f) 0)
   (else
    (fuck-up))))

(define (hexapawn-player->symbol player)
  (case player
    ((1)  'player1)
    ((-1) 'player2)
    (else (fuck-up))))

(define (hexapawn-piece-at i j pieces) (list-ref (list-ref pieces i) j))

(define (hexapawn-piece-at? i j pieces)
  (not (= 0 (list-ref (list-ref pieces i) j))))

(define (hexapawn-move-forward? element i j pieces)
  (and
   (hexapawn-on-board? (+ i element) pieces)
   (not (hexapawn-piece-at? (+ i element) j pieces))))

(define (hexapawn-move-backward? element i j pieces)
  (and
   (hexapawn-on-board? (- i element) pieces)
   (not (hexapawn-piece-at? (- i element) j pieces))))

(define (hexapawn-move-left? element i j pieces)
  (and
   (hexapawn-on-board? (- j 1) pieces)
   (not (hexapawn-piece-at? i (- j 1) pieces))))

(define (hexapawn-move-right? element i j pieces)
  (and
   (hexapawn-on-board? (+ j 1) pieces)
   (not (hexapawn-piece-at? i (+ j 1) pieces))))

(define (hexapawn-move-sideways? element direction i j pieces)
  (and
   (hexapawn-on-board? (+ i element) pieces)
   (hexapawn-on-board? (+ j direction) pieces)
   (not (hexapawn-piece-at? (+ i element) (+ j direction) pieces))))

(define (hexapawn-move-back-sideways? element direction i j pieces)
  (and
   (hexapawn-on-board? (- i element) pieces)
   (hexapawn-on-board? (+ j direction) pieces)
   (not (hexapawn-piece-at? (- i element) (+ j direction) pieces))))

(define (hexapawn-on-board? i pieces)
  (and (>= i 0)
       (< i (length pieces))))

(define (hexapawn-can-capture? element direction i j pieces)
  (and
   (hexapawn-on-board? (+ i element) pieces)
   (hexapawn-on-board? (+ j direction) pieces)
   (hexapawn-piece-at? (+ i element) (+ j direction) pieces)
   (not (eq?
	 (hexapawn-piece-at (+ i element) (+ j direction) pieces)
	 element))))

(define (hexapawn-can-capture-l? element i j pieces)
  (hexapawn-can-capture? element 1 i j pieces))

(define (hexapawn-can-capture-r? element i j pieces)
  (hexapawn-can-capture? element -1 i j pieces))

(define (hexapawn-moves board)
(case *variant*
  ((normal)
   (let ((pieces (board-playing-area board)))
     (apply
      append
      (map-indexed
      (lambda (row i)
	(apply
	 append
	 (map-indexed
	  (lambda (element j)
	    (if (= element (board-player board))
		(append
		 (if (hexapawn-move-forward? element i j pieces)
		     `((move (,i ,j) (,(+ i element) ,j) ,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-l? element i j pieces)
		     `((capture (,i ,j) (,(+ i element) ,(+ j 1))
				,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-r? element i j pieces)
		   `((capture (,i ,j) (,(+ i element) ,(- j 1))
			      ,(board-player board)))
		   '()))
		'()))
	  row)))
      pieces))))
  ((variant-a)
   (let ((pieces (board-playing-area board)))
     (apply
      append
      (map-indexed
       (lambda (row i)
	 (apply
	  append
	  (map-indexed
	   (lambda (element j)
	     (if (= element (board-player board))
		 (append
		  (if (hexapawn-move-sideways? element 1 i j pieces)
		      `((move (,i ,j) (,(+ i element) ,(+ j 1)) ,(board-player board)))
		      '())
		  (if (hexapawn-move-sideways? element -1 i j pieces)
		      `((move (,i ,j) (,(+ i element) ,(- j 1)) ,(board-player board)))
		      '())
		  (if (hexapawn-can-capture-l? element i j pieces)
		      `((capture (,i ,j) (,(+ i element) ,(+ j 1))
				 ,(board-player board)))
		      '())
		  (if (hexapawn-can-capture-r? element i j pieces)
		      `((capture (,i ,j) (,(+ i element) ,(- j 1))
				 ,(board-player board)))
		      '()))
		 '()))
	   row)))
       pieces))))
  ((variant-d)
   (let ((pieces (board-playing-area board)))
     (apply
      append
      (map-indexed
      (lambda (row i)
	(apply
	 append
	 (map-indexed
	  (lambda (element j)
	    (if (= element (board-player board))
		(append
		 (if (hexapawn-move-forward? element i j pieces)
		     `((move (,i ,j) (,(+ i element) ,j) ,(board-player board)))
		     '())
		 (if (hexapawn-move-backward? element i j pieces)
		     `((move (,i ,j) (,(- i element) ,j) ,(board-player board)))
		     '())
 		 (if (hexapawn-move-left? element i j pieces)
		     `((move (,i ,j) (,i ,(- j 1)) ,(board-player board)))
		     '())
		 (if (hexapawn-move-right? element i j pieces)
		     `((move (,i ,j) (,i ,(+ j 1)) ,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-l? element i j pieces)
		     `((capture (,i ,j) (,(+ i element) ,(+ j 1))
				,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-r? element i j pieces)
		   `((capture (,i ,j) (,(+ i element) ,(- j 1))
			      ,(board-player board)))
		   '()))
		'()))
	  row)))
      pieces))))
  ((variant-c)
   (let ((pieces (board-playing-area board)))
     (apply
      append
      (map-indexed
      (lambda (row i)
	(apply
	 append
	 (map-indexed
	  (lambda (element j)
	    (if (= element (board-player board))
		(append
		 (if (hexapawn-move-forward? element i j pieces)
		     `((move (,i ,j) (,(+ i element) ,j) ,(board-player board)))
		     '())
		 (if (hexapawn-move-backward? element i j pieces)
		     `((move (,i ,j) (,(- i element) ,j) ,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-l? element i j pieces)
		     `((capture (,i ,j) (,(+ i element) ,(+ j 1))
				,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-r? element i j pieces)
		   `((capture (,i ,j) (,(+ i element) ,(- j 1))
			      ,(board-player board)))
		   '()))
		'()))
	  row)))
      pieces))))
  ((variant-b)
   (let ((pieces (board-playing-area board)))
     (apply
      append
      (map-indexed
      (lambda (row i)
	(apply
	 append
	 (map-indexed
	  (lambda (element j)
	    (if (= element (board-player board))
		(append
		  (if (hexapawn-move-sideways? element 1 i j pieces)
		      `((move (,i ,j) (,(+ i element) ,(+ j 1)) ,(board-player board)))
		      '())
		  (if (hexapawn-move-sideways? element -1 i j pieces)
		      `((move (,i ,j) (,(+ i element) ,(- j 1)) ,(board-player board)))
		      '())
		  (if (hexapawn-move-back-sideways? element 1 i j pieces)
		      `((move (,i ,j) (,(- i element) ,(+ j 1)) ,(board-player board)))
		      '())
		  (if (hexapawn-move-back-sideways? element -1 i j pieces)
		      `((move (,i ,j) (,(- i element) ,(- j 1)) ,(board-player board)))
		      '())
		 (if (hexapawn-can-capture-l? element i j pieces)
		     `((capture (,i ,j) (,(+ i element) ,(+ j 1))
				,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-r? element i j pieces)
		   `((capture (,i ,j) (,(+ i element) ,(- j 1))
			      ,(board-player board)))
		   '()))
		'()))
	  row)))
      pieces))))))

(define hexapawn-move-type car)
(define hexapawn-move-from cadr)
(define hexapawn-move-to caddr)
(define hexapawn-move-player cadddr)
(define (hexapawn-move->player-type move)
  (simple-move->player move))
(define (hexapawn-move->piece-type move)
  (hexapawn-piece-num->piece-type (hexapawn-move-player move)))
(define hexapawn-row car)
(define hexapawn-column cadr)
(define (hexapawn-initial-player s) 1)

(define (simple-board-set-nth e i j board)
  (list
   (board-player board)
   (list-replace-ith
    (board-playing-area board)
    i
    (list-replace-ith
     (list-ref (board-playing-area board) i) j e))))

(define (hexapawn-make-move move board)
  (let ((from (hexapawn-move-from move))
	(to (hexapawn-move-to move)))
    (simple-board-set-nth
     0
     (hexapawn-row from)
     (hexapawn-column from)
     (simple-board-set-nth
      (hexapawn-piece-at (hexapawn-row from)
			 (hexapawn-column from)
			 (board-playing-area board))
      (hexapawn-row to)
      (hexapawn-column to)
      board))))

(define (hexapawn-win0 board)
  (cond
   ((some (lambda (x) (= -1 x)) (car (board-playing-area board)))
    -1)
   ((some (lambda (x) (= 1 x)) (last (board-playing-area board)))
    1)
   ((null? (moves board)) (* -1 (board-player board)))
   (else 0)))

(define (hexapawn-get-square-by-coord i j type squares)
  (find-if (lambda (square)
	     (and (eq? (first (square-name square)) type)
		  (eq? (second (square-name square)) i)
		  (eq? (third (square-name square)) j)))
	    squares))

(define (hexapawn-get-free-staging squares)
  (find-if (lambda (square)
	     (and (eq? (first (square-name square)) 'staging)
		  (not (square-contents square))))
	   *squares*))

(define (hexapawn-move-piece-action from to)
  (list (lambda () (pick-up
		    (hexapawn-get-square-by-coord
		     (cadr from)
		     (caddr from)
		     (car from)
		     *squares*)))
	(lambda () (put-down
		    (hexapawn-get-square-by-coord
		     (cadr to)
		     (caddr to)
		     (car to)
		     *squares*)))))

(define (hexapawn-move->actions player move)
  (case (car move)
    ((move)
     (hexapawn-move-piece-action `(board ,@(cadr move))
				 `(board ,@(caddr move))))
    ((capture)
     (let ((staging (hexapawn-get-free-staging *squares*)))
       (append
	(hexapawn-move-piece-action `(board ,@(hexapawn-move-to move))
				    (square-name staging))
	(hexapawn-move-piece-action `(board ,@(hexapawn-move-from move))
				    `(board ,@(hexapawn-move-to move))))))
    (else (fuck-up))))

(define (hexapawn-next-player board)
  (player-update
   board
   (lambda (board) (* -1 (board-player board)))))

(define (hexapawn-board->prolog board name) (simple-board->prolog board))

(define (hexapawn-piece-num->piece-type piece-num)
  (case piece-num
    ((1)  'x)
    ((0)  'none)
    ((-1) 'b)
    (else (fuck-up))))

(define (hexapawn-move->prolog move_ prev next)
  (let ((move (cadr move_)))
    `(("move" (,(symbol->lowercase-string (hexapawn-move->player-type move_))
	       ,(string-append
		 (symbol->lowercase-string (hexapawn-move->piece-type move))
		 "_act")
	       ,(number->row (hexapawn-row (hexapawn-move-from move)))
	       ,(number->column (hexapawn-column (hexapawn-move-from move)))
	       ,(number->row (hexapawn-row (hexapawn-move-to move)))
	       ,(number->column (hexapawn-column (hexapawn-move-to move)))
	       ,prev
	       ,next)))))

(define (hexapawn-board-outcome->prolog board) (panic "unimplemented"))

(define (hexapawn-game->prolog game) (simple-game->prolog game))

(define (hexapawn-flip-piece p)
  (cond
   ((not p) p)
   ((eq? p 'w) 'b)
   ((eq? p 'b) 'w)
   ((number? p) (* -1 p))
   (else (panic "Bad piece"))))
(define (hexapawn-flip-square s)
  (square-name-update hexapawn-flip-piece s))

(define (deep-map f l)
  (let ((new-l (f l)))
    (if (or (null? l) (not (list? new-l)))
	new-l
	(map (lambda (e) (deep-map f e)) new-l))))

(define (hexapawn-flip-player x) (* -1 x))

(define (hexapawn-flip-pieces game)
  (deep-map
   (lambda (x)
     (cond
      ((square? x) (hexapawn-flip-square x))
      ((and (list? x)
	    (not (null? x))
	    (eq? (car x) 'move)
	    (= (length x) 3))
       `(move ,(hexapawn-flip-player (cadr x)) ,(caddr x)))
      ((and (list? x)
	    (not (null? x))
	    (eq? (car x) 'move)
	    (= (length x) 4)
	    (list? (cadr x)))
       `(move ,(cadr x) ,(caddr x) ,(hexapawn-flip-piece (cadddr x))))
      (else x)))
   game))

(define (hexapawn-setup-new-game settings)
  (set! *game* 'hexapawn)
  (let ((size (car settings)))
    (set! *playing-area-size* size)
    (set! *squares* (hexapawn-generate-board size))))

(define (hexapawn-generate-board size)
  (simple-generate-board
   (lambda (r c)
     (cond ((= r 0) 'w)
	   ((= r (- size 1)) 'b)
	   (else #f)))
   size))

(define (simple-generate-board f size)
  (apply
   append
   (map-n
    (lambda (n)
      (if (< n 6)
	  (create-square 20 (+ 320 (* n 40)) #f `(staging left ,n))
	  (create-square 260 (+ 320 (* (- n 6) 40)) #f `(staging right ,(- n 6)))))
    (* 2 size))
   (map-n
    (lambda (c)
      (map-n
       (lambda (r)
	 (create-square (+ 100 (* c 40))
			(+ 360 (* r 40))
			(f r c)
			`(board ,r ,c)))
       size))
    size)))


;; N & C Snap

(define snap-utterances '("play" "colour" "shape" "same" "nothing"))
(define snap-shapes '("ring" "star" "flash"))
(define snap-colours '("red" "green" "blue"))
(define snap-positions '("pos0" "pos1"))

(define (snap-decls)
  (append
   (map (lambda (u) `("utterance" (,u))) snap-utterances)
   (map (lambda (u) `("shape" (,u))) snap-shapes)
   (map (lambda (u) `("colour" (,u))) snap-colours)
   (map (lambda (u) `("position" (,u))) snap-positions)))

(define (snap1-decls)
  (append
   (list (modeh '(1 ("action" "#utterance" "+time")))
	 (modeb '(* ("state" (list (list "-shape" "-colour" "-position")
				     (list "-shape" "-colour" "-position"))
		       "+time")))
	 (modeb '(* ("state" () "+time")))
	 (modeb '(* ("successor" "+time" "-time")))
	 (positive-only-learning))
   (snap-decls)))

(define (snap2-decls)
  (append
   (list (modeh '(1 ("action" "#utterance" "+time")))
	 (modeb '(* ("state" (list (list "-shape" "-colour" "-position")) "+time")))
	 (modeb '(* ("successor" "+time" "-time")))
	 (positive-only-learning))
   (snap-decls)))

(define (snap-action c1 c2)
  (cond ((and (null? c1) (null? c2)) "")
	((and (eq? (car c1) (car c2))
	      (eq? (cadr c1) (cadr c2))) "same")
	((eq? (car c1) (car c2)) "shape")
	((eq? (cadr c1) (cadr c2)) "colour")
	(else "nothing")))

(define (generate-filled-snap1 t-prev t-curr)
  (let ((c1 `(,(random-member snap-shapes)
	      ,(random-member snap-colours)
	      ,(car snap-positions)))
	(c2 `(,(random-member snap-shapes)
	      ,(random-member snap-colours)
	      ,(cadr snap-positions))))
    (snap1->prolog `(,c1 ,c2) `(,(snap-action c1 c2)) t-curr t-prev)))

(define (generate-empty-snap1 t-prev t-curr)
  (snap1->prolog '() '("play") t-curr t-prev))

(define (number->time n) (string-append "t" (number->string n)))

(define (generate-one-snap1 n)
  (append
   (generate-empty-snap1 (number->time n) (number->time (+ n 1)))
   (generate-filled-snap1 (number->time (+ n 1)) (number->time (+ n 2)))))

(define (generate-snap1 size)
  (prolog-program->string (generate-snap1-prolog size)))

(define (generate-snap1-prolog size)
  (snap1-decls)
  (apply append
	  (map-n (lambda (x) (generate-one-snap1 (* 2 x))) size)))

(define (snap2->prolog state action time succ)
  `(("state" ((list (list ,(car state) ,(cadr state) ,(caddr state)))
	      ,time))
    ,@(if (null? action)
	  '()
	  `(("action" (,(car action) ,time))))
    ("time" (,time))
    ("successor" (,succ ,time))))

(define (snap1->prolog state action time succ)
  `(("state" (,(if (null? state)
		   `()
		   `(list (list ,(caar state) ,(cadar state) ,(caddar state))
			  (list ,(caadr state) ,(cadadr state) ,(caddadr state))))
	      ,time))
    ,@(if (null? action)
	  '()
	  `(("action" (,(car action) ,time))))
    ("time" (,time))
    ("successor" (,succ ,time))))
