:- modeh(1,outcome(+board,+win_outcome))?

:- modeb(*,owns_piece(+win_outcome,-piece))?
:- modeb(*,at(-row,-col,+board,+piece))?
:- modeb(*,linear(+row,+col,+row,+col,+row,+col))?

:- set(noise,0)?
:- set(c,10)?
:- set(h,10000)?
:- set(r,40000)?
:- set(i,2000)?
:- set(n,2000)?
:- set(inflate,5000)?
:- set(nodes,4000)?
:- set(verbose,0)?

:- outcome(X,x_wins), outcome(X,o_wins).
:- outcome(X,x_wins), outcome(X,o_wins).

player(player_x).
player(player_o).
piece(x).
piece(o).
piece(none).
win_outcome(x_wins).
win_outcome(o_wins).
owns(player_x,x).
owns(player_o,o).
owns_outcome(player_x,x_wins).
owns_outcome(player_o,o_wins).
owns_piece(x_wins,x).
owns_piece(o_wins,o).
opponent(player_x,player_o).
opponent(player_o,player_x).

prune(_,Body) :- in(frame(A,B,C,D),Body), C==D.

%% The following constraints do not modify the outcome, they just
%% speed up search significantly
%% prune(_,Body) :-
%% setof(linear(A,B,C,D,E,F),in(linear(A,B,C,D,E,F),Body),[_,_|_]).

different(A,B) :- not(A = B).

linear_test(X1,Y1,X2,Y2,X3,Y3,S) :-
	S is X1*(Y2-Y3)+X2*(Y3-Y1)+X3*(Y1-Y2).
linear(R1,C1,R2,C2,R3,C3) :-  
	[R1,C1] \= [R2,C2],
	[R3,C3] \= [R2,C2],
	[R1,C1] \= [R3,C3],
	row_to_int(R1,RI1),
	row_to_int(R2,RI2),
	row_to_int(R3,RI3),
	col_to_int(C1,CI1),
	col_to_int(C2,CI2),
	col_to_int(C3,CI3),
	linear_test(RI1,CI1,RI2,CI2,RI3,CI3,0).



row_to_int(r0,0).
row_to_int(r1,1).
row_to_int(r2,2).
col_to_int(c0,0).
col_to_int(c1,1).
col_to_int(c2,2).

ref(0,[H|T],H).
ref(I,[H|T],H2) :- dec(I,J), ref(J,T,H2).

inc(X,Y) :- Y is X + 1.
dec(X,Y) :- Y is X - 1.

replace(0,H1,[H2|T],[H1|T]).
replace(I,H1,[H2|T1],[H2|T2]) :- dec(I,J), replace(J,H1,T1,T2).

at(ROW,COLUMN,BOARD,PIECE) :- row_to_int(ROW,RI),
	col_to_int(COLUMN,CI),
	ref(RI,BOARD,L),ref(CI,L,PIECE).

frame(RROW,CCOLUMN,BOARD1,BOARD2) :-
	row(RROW),
	col(CCOLUMN),
	row_to_int(RROW,ROW),
	col_to_int(CCOLUMN,COLUMN),
	board(BOARD1),
	board(BOARD2),
	ref(ROW,BOARD1,L1),
	replace(COLUMN,ignore,L1,L2),
	replace(ROW,L2,BOARD1,BOARD3),
	ref(ROW,BOARD2,L3),
	replace(COLUMN,ignore,L3,L4),
	replace(ROW,L4,BOARD2,BOARD4),
	BOARD3 == BOARD4.

row(r0).
col(c0).
row(r1).
col(c1).
row(r2).
col(c2).

board([[none,none,none],[none,none,none],[none,none,none]]).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],x_wins).
legal_move(player_x,[[none,none,none],[none,none,none],[none,none,none]],[[none,none,none],[x,none,none],[none,none,none]]).
board([[none,none,none],[x,none,none],[none,none,none]]).
:- outcome([[none,none,none],[x,none,none],[none,none,none]],o_wins).
:- outcome([[none,none,none],[x,none,none],[none,none,none]],x_wins).
legal_move(player_o,[[none,none,none],[x,none,none],[none,none,none]],[[none,none,none],[x,none,none],[none,o,none]]).
board([[none,none,none],[x,none,none],[none,o,none]]).
:- outcome([[none,none,none],[x,none,none],[none,o,none]],o_wins).
:- outcome([[none,none,none],[x,none,none],[none,o,none]],x_wins).
legal_move(player_x,[[none,none,none],[x,none,none],[none,o,none]],[[none,x,none],[x,none,none],[none,o,none]]).
board([[none,x,none],[x,none,none],[none,o,none]]).
:- outcome([[none,x,none],[x,none,none],[none,o,none]],o_wins).
:- outcome([[none,x,none],[x,none,none],[none,o,none]],x_wins).
legal_move(player_o,[[none,x,none],[x,none,none],[none,o,none]],[[none,x,none],[x,none,none],[o,o,none]]).
board([[none,x,none],[x,none,none],[o,o,none]]).
:- outcome([[none,x,none],[x,none,none],[o,o,none]],o_wins).
:- outcome([[none,x,none],[x,none,none],[o,o,none]],x_wins).
legal_move(player_x,[[none,x,none],[x,none,none],[o,o,none]],[[none,x,none],[x,x,none],[o,o,none]]).
board([[none,x,none],[x,x,none],[o,o,none]]).
:- outcome([[none,x,none],[x,x,none],[o,o,none]],o_wins).
:- outcome([[none,x,none],[x,x,none],[o,o,none]],x_wins).
legal_move(player_o,[[none,x,none],[x,x,none],[o,o,none]],[[none,x,none],[x,x,o],[o,o,none]]).
board([[none,x,none],[x,x,o],[o,o,none]]).
:- outcome([[none,x,none],[x,x,o],[o,o,none]],o_wins).
:- outcome([[none,x,none],[x,x,o],[o,o,none]],x_wins).
legal_move(player_x,[[none,x,none],[x,x,o],[o,o,none]],[[none,x,x],[x,x,o],[o,o,none]]).
board([[none,x,x],[x,x,o],[o,o,none]]).
:- outcome([[none,x,x],[x,x,o],[o,o,none]],o_wins).
:- outcome([[none,x,x],[x,x,o],[o,o,none]],x_wins).
legal_move(player_o,[[none,x,x],[x,x,o],[o,o,none]],[[none,x,x],[x,x,o],[o,o,o]]).
board([[none,x,x],[x,x,o],[o,o,o]]).
outcome([[none,x,x],[x,x,o],[o,o,o]],o_wins).
:- outcome([[none,x,x],[x,x,o],[o,o,o]],x_wins).


board([[none,none,none],[none,none,none],[none,none,none]]).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],x_wins).
legal_move(player_x,[[none,none,none],[none,none,none],[none,none,none]],[[none,none,none],[none,none,none],[none,x,none]]).
board([[none,none,none],[none,none,none],[none,x,none]]).
:- outcome([[none,none,none],[none,none,none],[none,x,none]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,x,none]],x_wins).
legal_move(player_o,[[none,none,none],[none,none,none],[none,x,none]],[[none,none,none],[none,none,o],[none,x,none]]).
board([[none,none,none],[none,none,o],[none,x,none]]).
:- outcome([[none,none,none],[none,none,o],[none,x,none]],o_wins).
:- outcome([[none,none,none],[none,none,o],[none,x,none]],x_wins).
legal_move(player_x,[[none,none,none],[none,none,o],[none,x,none]],[[none,none,x],[none,none,o],[none,x,none]]).
board([[none,none,x],[none,none,o],[none,x,none]]).
:- outcome([[none,none,x],[none,none,o],[none,x,none]],o_wins).
:- outcome([[none,none,x],[none,none,o],[none,x,none]],x_wins).
legal_move(player_o,[[none,none,x],[none,none,o],[none,x,none]],[[none,o,x],[none,none,o],[none,x,none]]).
board([[none,o,x],[none,none,o],[none,x,none]]).
:- outcome([[none,o,x],[none,none,o],[none,x,none]],o_wins).
:- outcome([[none,o,x],[none,none,o],[none,x,none]],x_wins).
legal_move(player_x,[[none,o,x],[none,none,o],[none,x,none]],[[x,o,x],[none,none,o],[none,x,none]]).
board([[x,o,x],[none,none,o],[none,x,none]]).
:- outcome([[x,o,x],[none,none,o],[none,x,none]],o_wins).
:- outcome([[x,o,x],[none,none,o],[none,x,none]],x_wins).
legal_move(player_o,[[x,o,x],[none,none,o],[none,x,none]],[[x,o,x],[o,none,o],[none,x,none]]).
board([[x,o,x],[o,none,o],[none,x,none]]).
:- outcome([[x,o,x],[o,none,o],[none,x,none]],o_wins).
:- outcome([[x,o,x],[o,none,o],[none,x,none]],x_wins).
legal_move(player_x,[[x,o,x],[o,none,o],[none,x,none]],[[x,o,x],[o,x,o],[none,x,none]]).
board([[x,o,x],[o,x,o],[none,x,none]]).
:- outcome([[x,o,x],[o,x,o],[none,x,none]],o_wins).
:- outcome([[x,o,x],[o,x,o],[none,x,none]],x_wins).
legal_move(player_o,[[x,o,x],[o,x,o],[none,x,none]],[[x,o,x],[o,x,o],[o,x,none]]).
board([[x,o,x],[o,x,o],[o,x,none]]).
:- outcome([[x,o,x],[o,x,o],[o,x,none]],o_wins).
:- outcome([[x,o,x],[o,x,o],[o,x,none]],x_wins).
legal_move(player_x,[[x,o,x],[o,x,o],[o,x,none]],[[x,o,x],[o,x,o],[o,x,x]]).
board([[x,o,x],[o,x,o],[o,x,x]]).
:- outcome([[x,o,x],[o,x,o],[o,x,x]],o_wins).
outcome([[x,o,x],[o,x,o],[o,x,x]],x_wins).

board([[none,none,none],[none,none,none],[none,none,none]]).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],x_wins).
legal_move(player_x,[[none,none,none],[none,none,none],[none,none,none]],[[none,none,none],[none,none,none],[none,none,x]]).
board([[none,none,none],[none,none,none],[none,none,x]]).
:- outcome([[none,none,none],[none,none,none],[none,none,x]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,none,x]],x_wins).
legal_move(player_o,[[none,none,none],[none,none,none],[none,none,x]],[[none,none,none],[none,none,none],[none,o,x]]).
board([[none,none,none],[none,none,none],[none,o,x]]).
:- outcome([[none,none,none],[none,none,none],[none,o,x]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,o,x]],x_wins).
legal_move(player_x,[[none,none,none],[none,none,none],[none,o,x]],[[none,x,none],[none,none,none],[none,o,x]]).
board([[none,x,none],[none,none,none],[none,o,x]]).
:- outcome([[none,x,none],[none,none,none],[none,o,x]],o_wins).
:- outcome([[none,x,none],[none,none,none],[none,o,x]],x_wins).
legal_move(player_o,[[none,x,none],[none,none,none],[none,o,x]],[[none,x,o],[none,none,none],[none,o,x]]).
board([[none,x,o],[none,none,none],[none,o,x]]).
:- outcome([[none,x,o],[none,none,none],[none,o,x]],o_wins).
:- outcome([[none,x,o],[none,none,none],[none,o,x]],x_wins).
legal_move(player_x,[[none,x,o],[none,none,none],[none,o,x]],[[none,x,o],[none,x,none],[none,o,x]]).
board([[none,x,o],[none,x,none],[none,o,x]]).
:- outcome([[none,x,o],[none,x,none],[none,o,x]],o_wins).
:- outcome([[none,x,o],[none,x,none],[none,o,x]],x_wins).
legal_move(player_o,[[none,x,o],[none,x,none],[none,o,x]],[[none,x,o],[none,x,none],[o,o,x]]).
board([[none,x,o],[none,x,none],[o,o,x]]).
:- outcome([[none,x,o],[none,x,none],[o,o,x]],o_wins).
:- outcome([[none,x,o],[none,x,none],[o,o,x]],x_wins).
legal_move(player_x,[[none,x,o],[none,x,none],[o,o,x]],[[none,x,o],[x,x,none],[o,o,x]]).
board([[none,x,o],[x,x,none],[o,o,x]]).
:- outcome([[none,x,o],[x,x,none],[o,o,x]],o_wins).
:- outcome([[none,x,o],[x,x,none],[o,o,x]],x_wins).
legal_move(player_o,[[none,x,o],[x,x,none],[o,o,x]],[[o,x,o],[x,x,none],[o,o,x]]).
board([[o,x,o],[x,x,none],[o,o,x]]).
:- outcome([[o,x,o],[x,x,none],[o,o,x]],o_wins).
:- outcome([[o,x,o],[x,x,none],[o,o,x]],x_wins).
legal_move(player_x,[[o,x,o],[x,x,none],[o,o,x]],[[o,x,o],[x,x,x],[o,o,x]]).
board([[o,x,o],[x,x,x],[o,o,x]]).
:- outcome([[o,x,o],[x,x,x],[o,o,x]],o_wins).
outcome([[o,x,o],[x,x,x],[o,o,x]],x_wins).

board([[none,none,none],[none,none,none],[none,none,none]]).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],o_wins).
:- outcome([[none,none,none],[none,none,none],[none,none,none]],x_wins).
legal_move(player_x,[[none,none,none],[none,none,none],[none,none,none]],[[x,none,none],[none,none,none],[none,none,none]]).
board([[x,none,none],[none,none,none],[none,none,none]]).
:- outcome([[x,none,none],[none,none,none],[none,none,none]],o_wins).
:- outcome([[x,none,none],[none,none,none],[none,none,none]],x_wins).
legal_move(player_o,[[x,none,none],[none,none,none],[none,none,none]],[[x,none,none],[o,none,none],[none,none,none]]).
board([[x,none,none],[o,none,none],[none,none,none]]).
:- outcome([[x,none,none],[o,none,none],[none,none,none]],o_wins).
:- outcome([[x,none,none],[o,none,none],[none,none,none]],x_wins).
legal_move(player_x,[[x,none,none],[o,none,none],[none,none,none]],[[x,none,none],[o,none,x],[none,none,none]]).
board([[x,none,none],[o,none,x],[none,none,none]]).
:- outcome([[x,none,none],[o,none,x],[none,none,none]],o_wins).
:- outcome([[x,none,none],[o,none,x],[none,none,none]],x_wins).
legal_move(player_o,[[x,none,none],[o,none,x],[none,none,none]],[[x,none,o],[o,none,x],[none,none,none]]).
board([[x,none,o],[o,none,x],[none,none,none]]).
:- outcome([[x,none,o],[o,none,x],[none,none,none]],o_wins).
:- outcome([[x,none,o],[o,none,x],[none,none,none]],x_wins).
legal_move(player_x,[[x,none,o],[o,none,x],[none,none,none]],[[x,none,o],[o,none,x],[none,x,none]]).
board([[x,none,o],[o,none,x],[none,x,none]]).
:- outcome([[x,none,o],[o,none,x],[none,x,none]],o_wins).
:- outcome([[x,none,o],[o,none,x],[none,x,none]],x_wins).
legal_move(player_o,[[x,none,o],[o,none,x],[none,x,none]],[[x,none,o],[o,o,x],[none,x,none]]).
board([[x,none,o],[o,o,x],[none,x,none]]).
:- outcome([[x,none,o],[o,o,x],[none,x,none]],o_wins).
:- outcome([[x,none,o],[o,o,x],[none,x,none]],x_wins).
legal_move(player_x,[[x,none,o],[o,o,x],[none,x,none]],[[x,x,o],[o,o,x],[none,x,none]]).
board([[x,x,o],[o,o,x],[none,x,none]]).
:- outcome([[x,x,o],[o,o,x],[none,x,none]],o_wins).
:- outcome([[x,x,o],[o,o,x],[none,x,none]],x_wins).
legal_move(player_o,[[x,x,o],[o,o,x],[none,x,none]],[[x,x,o],[o,o,x],[none,x,o]]).
board([[x,x,o],[o,o,x],[none,x,o]]).
:- outcome([[x,x,o],[o,o,x],[none,x,o]],o_wins).
:- outcome([[x,x,o],[o,o,x],[none,x,o]],x_wins).
legal_move(player_x,[[x,x,o],[o,o,x],[none,x,o]],[[x,x,o],[o,o,x],[x,x,o]]).
board([[x,x,o],[o,o,x],[x,x,o]]).
:- outcome([[x,x,o],[o,o,x],[x,x,o]],o_wins).
:- outcome([[x,x,o],[o,o,x],[x,x,o]],x_wins).

