(MODULE GAMES)

(include "QobiScheme.sch")
(include "c-externals.sc")

(define-external (pick-up x) TOOL1)
(define-external (put-down x) TOOL1)
(define-external (move-piece a b) TOOL1)

(define-structure square size x y contents name)
(define (create-square x y contents name) (make-square 40 x y contents name))
(define *squares* '())
(define *game* #f)


;;; Prolog

;;; clauses are represented as
;;; (name (head-vars) (call body))
;;; a rather literal translation of the scheme equivalent
;;; (clause->prolog-string '("tenu" ((list "D") () "N") (("digit" "D" "N"))))
;;; "tenu([D],[],N) :- digit(D,N)."

;;; TODO Andrei I abuse move to mean both the stored game one and the logical one

(define (caddadr x) (car (cddadr x)))

(define (symbol->lowercase-string s)
  (string-downcase (symbol->string s)))

(define (prolog-clause->string c)
  (if (and (string? (car c)) (null? (cdr c)))
      (car c)
      (string-append
       (if (null? (car c))
	   ""
	   (string-append
	    (prolog-clause-body->string (car c))
	    "("
	    (string-concat "," (map prolog-clause-body->string (cadr c)))
	    ")"))
       (if (or (null? (cddr c))
	       (null? (caddr c)))
	   ""
       (string-append
	" :- "
	(string-concat "," (map prolog-clause-body->string (caddr c)))))
       (if (equal? (last c) '?)
	   "?"
	   "."))))

(define (string-concat delim ss)
  (reduce (lambda (x y) (string-append x delim y)) ss ""))

(define (prolog-clause-body->string c)
  (cond ((null? c) "[]")
	((symbol? c) (symbol->string c))
	((string? c) c)
	((number? c) (number->string c))
	((or (equal? (car c) "*")
	     (equal? (car c) "+")
	     (equal? (car c) "-")
	     (equal? (car c) "/"))
	 (string-append
	  (prolog-clause-body->string (cadr c))
	  (car c)
	  (prolog-clause-body->string (caddr c))))
	((eq? (car c) 'list-pat)
	 (unless (> (length (cdr c)) 1) (panic "bad prolog atom ~a" c))
	 (string-append
	  "["
	  (let loop ((l (cddr c)) (s (prolog-clause-body->string (cadr c))))
	     (if (null? (cdr l))
		 (string-append
		  s
		  (string-append "|" (prolog-clause-body->string (car l))))
		 (loop (cdr l)
		       (string-append s "," (prolog-clause-body->string (car l))))))
	  "]"))
	((eq? (car c) 'list)
	 (string-append
	  "["
	  (string-concat
	   ","
	   (let loop ((l (cdr c)))
	     (if (null? l)
		 '()
		 (cons (prolog-clause-body->string (car l)) (loop (cdr l))))))
	  "]"))
	(else
	 (string-append
	  (prolog-clause-body->string (car c))
	  "("
	  (string-concat
	   ","
	   (map prolog-clause-body->string (cdr c)))
	  ")"))))

(define prolog->string prolog-clause->string)
(define (prolog-program->string p) (string-concat (list->string '(#\newline)) (map prolog->string p)))

(define (modeh b) `(() () (("modeh" ,@b)) ?))
(define (modeb b) `(() () (("modeb" ,@b)) ?))

(define (prolog-input x) (string-append "+" x))
(define (prolog-output x) (string-append "-" x))
(define (prolog-constant x) (string-append "#" x))

(define (prolog-set b) `(() () (("set" ,@b)) ?))

(define (positive-only-learning) (prolog-set '("posonly")))

(define *progol-binary* "progol")
(define *progol-output-file*
  (format "/tmp/progol-input-~a-~a.pl" (getenv "USERNAME") (getpid)))
(define *progol-input-file*
  (format "/tmp/progol-output-~a-~a.pl" (getenv "USERNAME") (getpid)))
(define (run-progol-and-learn progol-input)
  (system (format "rm -Rf ~a" *progol-output-file*))
  (system (format "rm -Rf ~a" *progol-input-file*))
  (write-file `(,progol-input) *progol-output-file*)
  (system (format "~a ~a > ~a"
		  *progol-binary*
		  *progol-output-file*
		  *progol-input-file*))
  (let ((result (read-file *progol-input-file*)))
    result))

(define (unlines lines)
  (reduce (lambda (x y) (string-append x (format "~%") y)) lines ""))

(define (char-downcase? c) (eq? (char-downcase c) c))

(define (prolog-clause-starts? line)
  (let ((unline (string->list line)))
    (and (not (null? unline))
	 (char-downcase? (car unline))
	 (char-alphabetic? (car unline)))))

(define (skip-until-prolog-clause lines_)
  (let loop ((lines lines_))
    (if (or
	 (null? lines)
	 (prolog-clause-starts? (car lines)))
	lines
	(loop (cdr lines)))))

(define (prolog-clause-ends? line)
  (findq #\. (string->list line)))

(define (prolog-join-clauses lines)
  (if (and (not (null? lines))
	   (prolog-clause-starts? (car lines))
	   (not (prolog-clause-ends? (car lines))))
      (prolog-join-clauses
       (cons (string-append (car lines)
			    (cadr lines))
	     (cddr lines)))
      lines))

(define (split-at x l)
  (let ((loc (position x l)))
    (if (or (null? l) (not loc))
	`(,l)
	(cons (sublist l 0 loc)
	      (split-at x (sublist l (+ loc 1) (length l)))))))

(define (split-string-at x s)
  (map list->string  (split-at x (string->list s))))

(define (split-clause line) (map (lambda (x)
				   (if (prefix? "-" x)
				       (substring x 1 (string-length x))
				       x))
				 (split-string-at #\: line)))

(define (string-position x s)
  (position x (string->list s)))

(define (prolog-read-end-marker line)
  `(,(substring line 0 (string-position #\. line)) "."))

(define (prolog-read-identifier lst)
  (let loop ((id '()) (l lst))
    (cond ((null? l)
	   `(,(list->string (reverse id)) ,l))
	  ((eq? #\( (car l))
	   (let ((x (prolog-read-args l)))
	     `((,(list->string (reverse id)) ,@(car x)) ,(cadr x))))
	  ((not (prolog-id-char? (car l)))
	   `(,(list->string (reverse id)) ,l))
	  (else
	   (loop (cons (car l) id) (cdr l))))))

(define (prolog-id-char? x)
  (or (char-alphanumeric? x)
      (eq? #\_ x)))

(define (prolog-read-args lst)
  (let loop ((ids '()) (l (if (eq? #\( (car lst))
			      (cdr lst)
			      lst)))
    (cond ((null? l)
	   `(,(reverse ids) ,l))
	  ((eq? #\) (car l))
	   `(,(reverse ids) ,(cdr l)))
	  ((eq? #\, (car l))
	   (loop ids (cdr l)))
	  ((char-whitespace? (car l))
	   (loop ids (cdr l)))
	  ((prolog-id-char? (car l))
	   (let ((x (prolog-read-identifier l)))
	     (loop (cons (car x) ids)  (cadr x))))
	  (else `(,(reverse ids) ,l)))))

(define (prolog-read-clause line)
  (let ((split (split-clause line)))
    (unless (= (length split) 2) #f)
    `(,@((lambda (x)
	   `(,(car x)
	     ,(cdr x)))
	 (car (prolog-read-identifier (string->list (car split)))))
      ,(car (prolog-read-args (string->list (cadr split)))))))

(define (read-prolog p)
  (let ((c (prolog-join-clauses (skip-until-prolog-clause p))))
    (if (null? c)
	'()
	(cons (prolog-read-clause (car c))
	      (read-prolog (cdr c))))))


;;; General interface
;;; generate-initial :: [parameters] -> (initial, start-player)
;;; generate-full-game :: initial -> start-player -> generator -> game
;;; generate-prolog :: game -> prolog-program
;;; So we generate a list of size two
;;; a representation of the actions for positive-only learning
;;; a representation of the outcomes for normal learning
;;; todo the ->prolog actions should be implemented

(define-structure board player playing-area)
(define (player-update board f)
  (make-board (f board) (board-playing-area board)))

(define (transpose-square s)
  (let* ((ss (copy-vector s))
	 (x (square-x ss))
	 (y (square-y ss)))
    (set-square-x! ss x)
    (set-square-y! ss y)
    ss))

(define (square-name-update f s)
  (let ((ns (copy-vector s)))
    (set-square-contents! ns (f (square-contents s)))
    ns))

(define *current-game* '())
(define *current-game-settings* #f)

;; TODO Andrei, we can actually calculate this
(define *playing-area-size* 3)

;; TODO Andrei, this should be either 0 or 100 by default, what makes more sense?
(define *percent-random-play* 100)

(define (copy-squares squares)
  (map (lambda (vector) (copy-vector vector)) squares))

(define (copy-vector vector) (list->vector (vector->list vector)))

(define (add-board! player area)
  (set! *current-game*
	(append *current-game*
		`(,(make-board player (copy-squares area))))))
(define (add-move! player move)
  (set! *current-game* (append *current-game* `((move ,player ,move)))))

(define (*squares*->board player)
  (make-board player (*squares*->playing-area)))

(define (squares->board squares player)
  (make-board player (squares->playing-area squares)))

(define (squares->playing-area squares)
  (map-n (lambda (i)
	   (map-n (lambda (j)
		    ((case *game*
		       ((xo) xo-get-square-contents)
		       ((hexapawn) hexapawn-get-square-contents)) i j squares))
		  *playing-area-size*))
	 *playing-area-size*))

(define (*squares*->playing-area) (squares->playing-area *squares*))

;; Game independent

(define (moves board)
 (case *game*
  ((xo) (xo-moves board))
  ((hexapawn) (hexapawn-moves board))
  (else (fuck-up))))

(define (make-move move board)
 (case *game*
  ((xo) (xo-make-move move board))
  ((hexapawn) (hexapawn-make-move move board))
  (else (fuck-up))))

(define (win0 board)
 (case *game*
  ((xo) (xo-win0 board))
  ((hexapawn) (hexapawn-win0 board))
  (else (fuck-up))))

(define (move->actions player move)
 (case *game*
  ((xo) (xo-move->actions player move))
  ((hexapawn) (hexapawn-move->actions player move))
  (else (fuck-up))))

(define (move->prolog move prev next)
 (case *game*
  ((xo) (xo-move->prolog move prev next))
  ((hexapawn) (hexapawn-move->prolog move prev next))
  (else (fuck-up))))

(define (maximize f l limit)
 (define (loop best-so-far l)
  (cond ((>= best-so-far limit) 1)
	((null? l) best-so-far)
	(else (loop (max (f (first l) best-so-far) best-so-far) (rest l)))))
 (loop -1 l))

(define (win*-alpha/beta board limit)
  (let ((player-board (board-player board))
	(win0-board (win0 board)))
    (cond ((not (zero? win0-board)) win0-board)
	  ((null? (moves board)) 0)
	  (else
	   (* player-board
	      (maximize
	       (lambda (move limit)
		 (* player-board
		    (win*-alpha/beta
		     (make-move move board) (* player-board limit))))
	       (moves board)
	       (* player-board limit)))))))

(define (win* board) (win*-alpha/beta board (board-player board)))

(define (optimal-moves board)
  (let ((board-winner (win* board))
	(player (board-player board)))
    (remove-if-not (lambda (move)
		     (= (* player (win* (make-move move board)))
			(* player board-winner)))
		   (moves board))))

(define (draw? board) (null? (moves board)))

(define (play! player)
 (let ((board (*squares*->board player)))
  (when (or (not (zero? (win0 board))) (null? (moves board)))
	(display "Game over")(newline)
   (message "Game over") (abort))
  (for-each (lambda (action) (action))
	    (begin
	      (let ((move (if (> *percent-random-play* (random-integer 99))
			      (random-member (moves board))
			      (random-member (optimal-moves board)))))
		(add-move! player move)
		(move->actions player move))))
  (add-board! player *squares*)))

(define (game->prolog game name)
  (case *game*
    ((xo) (xo-game->prolog game name))
    ((hexapawn) (hexapawn-game->prolog game name))))

(define (board->prolog board name)
  (case *game*
    ((xo) (xo-board->prolog board name))
    ((hexapawn) (hexapawn-board->prolog board name))))

(define (simple-prolog-negative-outcomes board name)
  (append
   (map
    (lambda (player)
      `(() () ("outcome" (win ,player))))
    (game-board-players board))
   `((() () ("outcome" draw)))))

(define (cadaddr x) (cadr (caddr x)))
(define (simple-prolog->outcome outcome) (cadaddr outcome))

(define (simple-outcome->prolog board name)
  (let ((outcome (win0 board))
	(moves (moves board)))
    (if (and (zero? outcome) (> (length moves) 0))
	(simple-prolog-negative-outcomes board name)
	(let ((sym-outcome (if (and (zero? outcome) (null? moves))
			       'draw
			       `(win ,(game-player->symbol outcome)))))
	  `(("outcome"
	     ,name
	     ,sym-outcome)
	    ,@(remove-if
	       (lambda (e)
		 (equal? (simple-prolog->outcome e) sym-outcome))
	       (simple-prolog-negative-outcomes board name)))))))

(define (next-player board)
  (case *game*
    ((xo) (xo-next-player board))
    ((hexapawn) (hexapawn-next-player board))
    (else (fuck-up))))

; (generate-game '(4))(game->prolog *current-game* "gen")
(define (generate-game settings)
  (setup-new-game settings)
  (call-with-current-continuation
   (lambda (c)
     (set! *abort-continuation* c)
     (let loop () (play! 1) (play! -1) (loop))))
  *current-game*)

(define (generate-board-prolog size)
  (apply append
	 (map-n (lambda (n)
		  `(("row" (,(string-append "r" (number->string n))))
		    ("col" (,(string-append "c" (number->string n))))))
		size)))

(define (restart-game) (setup-new-game *current-game-settings*))

(define (setup-new-game settings)
  (set! *current-game* '())
  (set! *current-game-settings* settings)
  (case *game*
    ((xo) (xo-setup-new-game settings)
     (add-board! (xo-initial-player settings) *squares* ))
    ((hexapawn)
     (hexapawn-setup-new-game settings)
     (add-board! (hexapawn-initial-player settings) *squares*))
    (else (fuck-up))))

(define (board-num->string name num) (string-append name (number->string num)))

;; XO

(define (xo-initial-player s) 1)
 (define (xo-get-square-contents i j squares)
  (case (square-contents
	 (find-if (lambda (square)
		   (and (eq? (first (square-name square)) 'board)
			(= i (second (square-name square)))
			(= j (third (square-name square)))))
		  squares))
   ((x) 1)
   ((o) -1)
   ((#f) 0)
   (else (fuck-up))))

(define (xo-setup-new-game settings)
  (set! *game* 'xo)
  (let ((size (car settings)))
    (unless (= size 3)
	    (message "tic-tac-toe only supports games of size 3")
	    (abort))
    (set! *playing-area-size* 3)
    (set! *squares*
	  (list
	   ;; Left staging area
	   (create-square 20 320 'x '(staging left 0))
	   (create-square 20 360 'x '(staging left 1))
	   (create-square 20 400 'x '(staging left 2))
	   (create-square 20 440 'x '(staging left 3))
	   (create-square 20 480 'x '(staging left 4))
	   ;; 3*3 board
	   (create-square 100 360 #f '(board 0 0))
	   (create-square 100 400 #f '(board 1 0))
	   (create-square 100 440 #f '(board 2 0))
	   (create-square 140 360 #f '(board 0 1))
	   (create-square 140 400 #f '(board 1 1))
	   (create-square 140 440 #f '(board 2 1))
	   (create-square 180 360 #f '(board 0 2))
	   (create-square 180 400 #f '(board 1 2))
	   (create-square 180 440 #f '(board 2 2))
	   ;; Right staging area
	   (create-square 260 320 'o '(staging right 0))
	   (create-square 260 360 'o '(staging right 1))
	   (create-square 260 400 'o '(staging right 2))
	   (create-square 260 440 'o '(staging right 3))
	   (create-square 260 480 'o '(staging right 4))))))

(define (simple-game->prolog game name)
  (let loop ((game game) (prolog '()) (board-num 0))
    (cond
     ((null? game) prolog)
     ((board? (car game))
      (loop
       (cdr game)
       (append
	prolog
	(board->prolog (car game)
		       (board-num->string name board-num)))
       (+ 1 board-num)))
     ((eq? (caar game) 'move)
      (loop
       (cdr game)
       (append
	prolog
	(move->prolog (cdar game)
			 (board-num->string name (- board-num 1))
			 (board-num->string name board-num)))
       board-num))
     (else (fuck-up)))))


(define xo-game->prolog simple-game->prolog)

(define (simple-board->prolog board name piecef)
  (append
   `(("board" (,name)))
   (simple-playing-positions->prolog
    (squares->playing-area (board-playing-area board))
    name
    piecef)))
;;;   (simple-outcome->prolog
;;;    board
;;;    name)))

(define (xo-board->prolog board name)
  (simple-board->prolog board name xo-piece-num->piece-type))

(define (simple-playing-positions->prolog playing-area name piecef)
  (apply
   append
   (map-indexed
    (lambda (row row-num)
      (map-indexed (lambda (entry col-num)
		     `("board_p"
		       (,(symbol->lowercase-string (piecef entry))
			,(number->row row-num)
			,(number->column col-num)
			,name)))
		   row))
    playing-area)))

(define (xo-move->prolog move prev next)
  `(("move" (,(symbol->lowercase-string (simple-move->player move))
	     ,(string-append
	       (symbol->lowercase-string (xo-move->piece-type move))
	       "_act")
	     ,(number->row (xo-move->row move))
	     ,(number->column (xo-move->column move))
	     ,prev
	     ,next))))

(define (number->column n) (format "c~a" n))
(define (number->row n) (format "r~a" n))

(define (xo-player->symbol player)
  (case player
    ((1)  'player1)
    ((-1) 'player2)
    (else (fuck-up))))

(define (simple-move->player move)
  (case (first move)
    ((1)  'player1)
    ((-1) 'player2)
    (else (fuck-up))))

(define (xo-move->piece-type move)
  (xo-piece-num->piece-type (first move)))

(define (xo-piece-num->piece-type piece-num)
  (case piece-num
    ((1)  'x)
    ((0)  'none)
    ((-1) 'o)
    (else (fuck-up))))

(define (xo-move->column move) (cadadr move))
(define (xo-move->row move) (caadr move))

(define (game-progol-prelude)
  (case *game*
    ((xo) (xo-progol-prelude))
    ((hexapawn) (hexapawn-progol-prelude))
    (else (fuck-up))))

(define (game-player->symbol player)
  (case *game*
    ((xo) (xo-player->symbol player))
    ((hexapawn) (hexapawn-player->symbol player))
    (else (fuck-up))))

(define (game-board-players board)
  (case *game*
    ((xo) (xo-game-board-players board))
    ((hexapawn) (hexapawn-game-board-players board))
    (else (fuck-up))))

(define (xo-game-board-players game) '(player1 player2))
(define (hexapawn-game-board-players game) '(player1 player2))

(define (game-prolog->progol prolog)
  (case *game*
    ((xo) (append
	   (xo-progol-prelude)
	   prolog))
    ((hexapawn)
     (append (hexapawn-progol-prelude)
	     prolog))
    (else (fuck-up))))

(define (game->progol game name)
  (game-prolog->progol (game->prolog game name)))

(define (learn-game game name)
  (read-prolog
   (cprogol-skip-to-results
    (run-progol-and-learn
     (prolog-program->string (game->progol game name))))))

(define (generate-games settings)
  (map (lambda (setting)
	 (generate-game setting)
	 *current-game*) settings))

(define (games->progol games names)
  (game-prolog->progol
   (apply
    append
    (map (lambda (e)
	   (game->prolog (first e)
			 (second e))) (zip games names)))))

(define (learn-games games names)
  (read-prolog
   (cprogol-skip-to-results
    (run-progol-and-learn
     (prolog-program->string
      (games->progol games names))))))


(define (cprogol-skip-to-results lines)
  (let ((lines (sublist lines (+ (position "" lines) 1) (length lines))))
    (sublist lines (position "" lines) (length lines))))

(define correct-output
  '(("move" ("A" "x_act" "B" "C" "D" "E") (("board" "E") ("board_p" "none" "B" "C" "D") ("board_p" "x" "B" "C" "E")))
    ("move" ("A" "o_act" "B" "C" "D" "E") (("board" "E") ("board_p" "none" "B" "C" "D") ("board_p" "o" "B" "C" "E")))))
;  "move(A,x_act,B,C,D,E) :- board(E),board_p(none,B,C,D),board_p(x,B,C,E).
;move(A,o_act,B,C,D,E) :- board(E),board_p(none,B,C,D),board_p(o,B,C,E).")

(define (generate-and-learn nr initial)
  (learn-games
   (generate-games (map-n (lambda (n) initial) nr))
   (map-n (lambda (n) (format "gen_~a_" n)) nr)))

(define (test-n-games nr g correct)
  (map-n (lambda (n) (let ((r (g)))
		       (if (equal? r correct)
			   #t
			   (begin (display r)(newline)
				  (display correct)
				  (newline))))) nr))

(define (test-n-generated-games nr gen-nr gen-initial correct)
  (test-n-games nr
		(lambda () (generate-and-learn gen-nr gen-initial))
		correct))

(define (game-statistics nr gen-nr gen-initial correct)
  (/
   (apply +
	  (map
	   (lambda (r) (if r 1 0))
	   (test-n-generated-games nr gen-nr gen-initial correct)))
   nr))

(define (zip a b)
  (if (or (null? a) (null? b))
      '()
      (cons (list (car a) (car b)) (zip (cdr a) (cdr b)))))

(define (xo-prolog->progol x) x)


(define (xo-moves board)
 (define (discard-non-empty row)
  (cond ((null? row) '())
	((zero? (third (first row)))
	 (cons (first row) (discard-non-empty (rest row))))
	(else (discard-non-empty (rest row)))))
 (map (lambda (move) (list (first move) (second move)))
      (reduce append
	      (map-indexed
	       (lambda (row i)
		 (discard-non-empty
		  (map-indexed (lambda (element j) (list i j element)) row)))
	       (board-playing-area board))
	      '())))

(define (list-replace-ith list i x)
  (if (= i 0)
      (cons x (rest list))
      (cons (first list) (list-replace-ith (rest list) (- i 1) x))))

(define (xo-make-move move board)
  (make-board
   (xo-next-player board)
   (list-replace-ith
    (board-playing-area board)
    (first move)
    (list-replace-ith
     (list-ref (board-playing-area board) (first move)) (second move) (board-player board)))))

(define (xo-win0 board)
 (define (at board i j) (list-ref (list-ref (board-playing-area board) i) j))
 (define (win? player)
  (let ((n (length (board-playing-area board))))
   (or
    (some-n (lambda (i) (every-n (lambda (j) (= (at board i j) player)) n)) n)
    (some-n (lambda (j) (every-n (lambda (i) (= (at board i j) player)) n)) n)
    (every-n (lambda (i) (= (at board i i) player)) n)
    (every-n (lambda (i) (= (at board i (- (- n 1) i)) player)) n))))
 (cond ((win? 1) 1)
       ((win? -1) -1)
       (else 0)))

(define (xo-move->actions player move)
 (let ((from (find-if (lambda (square)
		       (and (eq? (first (square-name square)) 'staging)
			    (eq? (square-contents square)
				 (if (= player 1) 'x 'o))))
		      *squares*))
       (to (find-if (lambda (square)
		     (and (eq? (first (square-name square)) 'board)
			  (= (second (square-name square)) (first move))
			  (= (third (square-name square)) (second move))))
		    *squares*)))
   (list (lambda () (pick-up from)) 
	 (lambda () (put-down to)))))

(define (xo-next-player board)
  (player-update
   board
   (lambda (board) (- (board-player board)))))

(define (xo-progol-prelude)
  `(,(modeh `(1 ("move"
		 ,(prolog-input "player")
		 ,(prolog-constant "act")
		 ,(prolog-input "row")
		 ,(prolog-input "col")
		 ,(prolog-input "board")
		 ,(prolog-output "board"))))
    ,(modeb `(* ("board" ,(prolog-input "board"))))
    ,(modeb `(* ("board" ,(prolog-output "board"))))
    ,(modeb `(* ("board_p"
		 ,(prolog-constant "piece")
		 ,(prolog-input "row")
		 ,(prolog-input "col")
		 ,(prolog-input "board"))))
    (":- modeb(*, linear([-loc, -loc], [-loc, -loc], [-loc, -loc]))?")
    ,(positive-only-learning)
    ,(prolog-set `("noise" 0))
    ,(prolog-set `("c" 3))
    ,(prolog-set `("h" 20))
    ,(prolog-set `("n" 50))
    ,(prolog-set `("r" 300))
    ,(prolog-set `("i" 10))
    ,(prolog-set `("inflate" 5000))
    ,(prolog-set `("nodes" 500))
    ,(prolog-set `("verbose" 0))
    ("player" ("player1"))
    ("player" ("player2"))
    ,@(generate-board-prolog *playing-area-size*)
    ("act" ("x_act"))
    ("act" ("o_act"))
    ("piece" ("x"))
    ("piece" ("o"))
    ("piece" ("none"))
    ("linear([A,B],[C,D],[E,F]) :- S is A/B, C is D*S, E is F*S.")))


(define correct-xo-output '("CProgol Version 4.4" "" "[:- modeh(1,move(+player,#act,+row,+col,+board,-board))? - Time taken 0.00s]" "[:- modeb(100,board(+board))? - Time taken 0.00s]" "[:- modeb(100,board(-board))? - Time taken 0.00s]" "[:- modeb(100,board_p(#piece,+row,+col,+board))? - Time taken 0.00s]" "[:- modeb(100,linear([-loc,-loc],[-loc,-loc],[-loc,-loc]))? - Time taken 0.00s]" "[Noise has been set to 100%]" "[Example inflation has been set to 400%]" "[The posonly flag has been turned ON]" "[:- set(posonly)? - Time taken 0.00s]" "[:- set(noise,0)? - Time taken 0.00s]" "[:- set(c,3)? - Time taken 0.00s]" "[:- set(h,20)? - Time taken 0.00s]" "[:- set(n,50)? - Time taken 0.00s]" "[:- set(r,300)? - Time taken 0.00s]" "[:- set(i,10)? - Time taken 0.00s]" "[:- set(nodes,500)? - Time taken 0.00s]" "[Generalising move(player1,x_act,r2,c2,gen_0_0,gen_0_1).]" "Resource limit exceeded" "[Generalising move(player2,o_act,r0,c1,gen_0_1,gen_0_2).]" "Resource limit exceeded" "" "move(A,x_act,B,C,D,E) :- board(E), board_p(none,B,C,D), board_p(x," "	B,C,E)." "move(A,o_act,B,C,D,E) :- board(E), board_p(o,B,C,E)." "" "[Total number of clauses = 2]" ""))




;;; Game Independent

;; ;; TODO
;; ;; Augment action with a player predicate
;; ;; player(player, board).

;; ;; player(p0, b0).
;; ;; player(p1, b0).
;; ;; action(..).

;; ;; One classification of game complexity should be on the complexity of taking turns.
;; ;; Independent or dependent turn games
;; ;; In the real world everything is independent, but that's quite a bit harder
;; ;; Neither xo nor hexapawn is indep. but I should still try it out.

;; ;; Can progol do predicate invention?

;; ;; Initial is
;; ;; initial(B) :- board(B), action(x,y,B,X), X not B.
;; ;; Is this general enough (are there games that can have different initials?)
;; ;; I should put this as a derivation rule for hypothesis instead of as B?
;; ;; Perhaps just learning initial(#board) will be enough

;; ;; I'm not going to include the concept of player? at a particular board position
;; ;; Well, I'll have to enhance the board definition with progol names, or generate thunks..


;; Hexapawn

(define (hexapawn-get-square-contents i j squares)
  (case (square-contents
	 (find-if (lambda (square)
		   (and (eq? (first (square-name square)) 'board)
			(= i (second (square-name square)))
			(= j (third (square-name square)))))
		  squares))
   ((w) 1)
   ((b) -1)
   ((#f) 0)
   (else
    (fuck-up))))

(define (hexapawn-player->symbol player)
  (case player
    ((1)  'player1)
    ((-1) 'player2)
    (else (fuck-up))))

(define (hexapawn-piece-at i j pieces) (list-ref (list-ref pieces i) j))

(define (hexapawn-piece-at? i j pieces)
  (not (= 0 (list-ref (list-ref pieces i) j))))

(define (hexapawn-move-forward? element i j pieces)
  (and
   (hexapawn-on-board? (+ i element) pieces)
   (not (hexapawn-piece-at? (+ i element) j pieces))))

(define (hexapawn-on-board? i pieces)
  (and (>= i 0)
       (< i (length pieces))))

(define (hexapawn-can-capture? element direction i j pieces)
  (and
   (hexapawn-on-board? (+ i element) pieces)
   (hexapawn-on-board? (+ j direction) pieces)
   (hexapawn-piece-at? (+ i element) (+ j direction) pieces)
   (not (eq?
	 (hexapawn-piece-at (+ i element) (+ j direction) pieces)
	 element))))

(define (hexapawn-can-capture-l? element i j pieces)
  (hexapawn-can-capture? element 1 i j pieces))

(define (hexapawn-can-capture-r? element i j pieces)
  (hexapawn-can-capture? element -1 i j pieces))

(define (hexapawn-moves board)
  (let ((pieces (board-playing-area board)))
    (apply
     append
     (map-indexed
      (lambda (row i)
	(apply
	 append
	 (map-indexed
	  (lambda (element j)
	    (if (= element (board-player board))
		(append
		 (if (hexapawn-move-forward? element i j pieces)
		     `((move (,i ,j) (,(+ i element) ,j) ,(board-player board)))
		     '())
		 (if (hexapawn-can-capture-l? element i j pieces)
		     `((capture (,i ,j) (,(+ i element) ,(+ j 1))
				,(board-player board)))
		     '())
	       (if (hexapawn-can-capture-r? element i j pieces)
		   `((capture (,i ,j) (,(+ i element) ,(- j 1))
			      ,(board-player board)))
		  '()))
		'()))
	  row)))
      pieces))))

(define hexapawn-move-type car)
(define hexapawn-move-from cadr)
(define hexapawn-move-to caddr)
(define hexapawn-move-player cadddr)
(define (hexapawn-move->player-type move)
  (simple-move->player move))
(define (hexapawn-move->piece-type move)
  (hexapawn-piece-num->piece-type (hexapawn-move-player move)))
(define hexapawn-row car)
(define hexapawn-column cadr)
(define (hexapawn-initial-player s) 1)

(define (simple-board-set-nth e i j board)
  (list
   (board-player board)
   (list-replace-ith
    (board-playing-area board)
    i
    (list-replace-ith
     (list-ref (board-playing-area board) i) j e))))

(define (hexapawn-make-move move board)
  (let ((from (hexapawn-move-from move))
	(to (hexapawn-move-to move)))
    (simple-board-set-nth
     0
     (hexapawn-row from)
     (hexapawn-column from)
     (simple-board-set-nth
      (hexapawn-piece-at (hexapawn-row from)
			 (hexapawn-column from)
			 (board-playing-area board))
      (hexapawn-row to)
      (hexapawn-column to)
      board))))

(define (hexapawn-win0 board)
  (cond
   ((some (lambda (x) (= -1 x)) (car (board-playing-area board)))
    -1)
   ((some (lambda (x) (= 1 x)) (last (board-playing-area board)))
    1)
   (else 0)))

(define (hexapawn-get-square-by-coord i j type squares)
  (find-if (lambda (square)
	     (and (eq? (first (square-name square)) type)
		  (eq? (second (square-name square)) i)
		  (eq? (third (square-name square)) j)))
	    squares))

(define (hexapawn-get-free-staging squares)
  (find-if (lambda (square)
	     (and (eq? (first (square-name square)) 'staging)
		  (not (square-contents square))))
	   *squares*))

(define (hexapawn-move-piece-action from to)
  (list (lambda () (move-piece 
		    'pick-up
		    (hexapawn-get-square-by-coord
		     (cadr from)
		     (caddr from)
		     (car from)
		     *squares*)))
	(lambda () (move-piece
		    'put-down
		    (hexapawn-get-square-by-coord
		     (cadr to)
		     (caddr to)
		     (car to)
		     *squares*)))))

(define (hexapawn-move->actions player move)
  (case (car move)
    ((move)
     (hexapawn-move-piece-action `(board ,@(cadr move))
				 `(board ,@(caddr move))))
    ((capture)
     (let ((staging (hexapawn-get-free-staging *squares*)))
       (append
	(hexapawn-move-piece-action `(board ,@(hexapawn-move-to move))
				    (square-name staging))
	(hexapawn-move-piece-action `(board ,@(hexapawn-move-from move))
				    `(board ,@(hexapawn-move-to move))))))
    (else (fuck-up))))

(define (hexapawn-next-player board)
  (player-update
   board
   (lambda (board) (* -1 (board-player board)))))

(define (hexapawn-board->prolog board name)
  (simple-board->prolog board name hexapawn-piece-num->piece-type))

(define (hexapawn-piece-num->piece-type piece-num)
  (case piece-num
    ((1)  'w)
    ((0)  'none)
    ((-1) 'b)
    (else (fuck-up))))

(define (hexapawn-move->prolog move_ prev next)
  (let ((move (cadr move_)))
    `(("move" (,(symbol->lowercase-string (hexapawn-move->player-type move_))
	       ,(string-append
		 (symbol->lowercase-string (hexapawn-move->piece-type move))
		 "_act")
	       ,(number->row (hexapawn-row (hexapawn-move-from move)))
	       ,(number->column (hexapawn-column (hexapawn-move-from move)))
	       ,(number->row (hexapawn-row (hexapawn-move-to move)))
	       ,(number->column (hexapawn-column (hexapawn-move-to move)))
	       ,prev
	       ,next)))))

(define (hexapawn-board-outcome->prolog board) (panic "unimplemented"))

(define (hexapawn-game->prolog game name)
  `(,@(simple-game->prolog game name)
    ,@(simple-game->prolog (hexapawn-flip-pieces game)
			   (string-append "symm_" name))))

(define (hexapawn-flip-piece p)
  (cond
   ((not p) p)
   ((eq? p 'w) 'b)
   ((eq? p 'b) 'w)
   ((number? p) (* -1 p))
   (else (panic "Bad piece"))))
(define (hexapawn-flip-square s)
  (square-name-update hexapawn-flip-piece s))

(define (deep-map f l)
  (let ((new-l (f l)))
    (if (or (null? l) (not (list? new-l)))
	new-l
	(map (lambda (e) (deep-map f e)) new-l))))

(define (hexapawn-flip-player x) (* -1 x))

(define (hexapawn-flip-pieces game)
  (deep-map
   (lambda (x)
     (cond
      ((square? x) (hexapawn-flip-square x))
      ((and (list? x)
	    (not (null? x))
	    (eq? (car x) 'move)
	    (= (length x) 3))
       `(move ,(hexapawn-flip-player (cadr x)) ,(caddr x)))
      ((and (list? x)
	    (not (null? x))
	    (eq? (car x) 'move)
	    (= (length x) 4)
	    (list? (cadr x)))
       `(move ,(cadr x) ,(caddr x) ,(hexapawn-flip-piece (cadddr x))))
      (else x)))
   game))

(define (hexapawn-transpose-game game)
  (deep-map
   (lambda (x)
     (cond
      ((square? x) (transpose-square x))
      ((and (list? x)
	    (not (null? x))
	    (eq? (car x) 'move)
	    (= (length x) 3))
       `(move ,(hexapawn-flip-player (cadr x))
	      ,(reverse (caddr x))))
      ((and (list? x)
	    (not (null? x))
	    (eq? (car x) 'move)
	    (= (length x) 4)
	    (list? (cadr x)))
       `(move ,(reverse (cadr x))
	      ,(reverse (caddr x))
	      ,(hexapawn-flip-piece (cadddr x))))
      (else x)))
   game))


(define (hexapawn-progol-prelude)
  `(,(modeh `(1 ("move"
		 ,(prolog-input "player")
		 ,(prolog-constant "act")
		 ,(prolog-input "row")
		 ,(prolog-input "col")
		 ,(prolog-input "row")
		 ,(prolog-input "col")
		 ,(prolog-input "board")
		 ,(prolog-output "board"))))
    ,(modeb `(* ("board" ,(prolog-input "board"))))
    ,(modeb `(* ("board" ,(prolog-output "board"))))
    ,(modeb `(* ("board_p"
		 ,(prolog-constant "piece")
		 ,(prolog-input "row")
		 ,(prolog-input "col")
		 ,(prolog-input "board"))))
    (":- modeb(*,row_diff(+row,+row,-drank))?
 :- modeb(*,row_diff(+row,+row,#drank))?
 :- modeb(*,row_diff(+row,-row,#drank))?
 :- modeb(*,col_diff(+col,+col,-drank))?
 :- modeb(*,col_diff(+col,+col,#drank))?
 :- modeb(*,col_diff(+col,-col,#drank))?")
    ,(positive-only-learning)
    ,(prolog-set `("noise" 0))
    ,(prolog-set `("c" 3))
    ,(prolog-set `("h" 20))
    ,(prolog-set `("n" 30))
    ,(prolog-set `("r" 300))
    ,(prolog-set `("i" 30))
    ,(prolog-set `("inflate" 500))
    ,(prolog-set `("nodes" 500))
    ,(prolog-set `("verbose" 0))
    ("player" ("player1"))
    ("player" ("player2"))
    ("row_diff(R1,R2,Z) :- row(R1), row(R2), toNum(R1,X), toNum(R2,Y), Z is X - Y.
col_diff(C1,C2,Z) :- col(C1), col(C2), toNum(C1,X), toNum(C2,Y), Z is X - Y.")
    ("adjacent(R1,C1,R2,C2) :- toNum(R1,NR1), toNum(C1,NC1), toNum(R2,NR2), toNum(C2,NC2), CD is NC1 - NC2, abs(CD, CA), RD is NR1 - NR2, abs(RD, RA), 1 >= CA, 1 >= RA.")
    ,@(generate-board-prolog *playing-area-size*)
    ("act" ("w_act"))
    ("act" ("b_act"))
    ,@(hexapawn-generate-pieces *playing-area-size*)
    ("piece" ("none"))))

(define (hexapawn-generate-pieces size)
  (apply
   append
   (map-n
    (lambda (n)
      `(("piece" (,(string-append "w" (number->string n))))
	("piece" (,(string-append "b" (number->string n))))))
    size)))


(define (hexapawn-setup-new-game settings)
  (set! *game* 'hexapawn)
  (let ((size (car settings)))
    (set! *playing-area-size* size)
    (set! *squares* (hexapawn-generate-board size))))

(define (hexapawn-generate-board size)
  (simple-generate-board
   (lambda (r c)
     (cond ((= r 0) 'w)
	   ((= r (- size 1)) 'b)
	   (else #f)))
   size))

(define (simple-generate-board f size)
  (apply
   append
   (map-n
    (lambda (n)
      (if (< n 6)
	  (create-square 20 (+ 320 (* n 40)) #f `(staging left ,n))
	  (create-square 260 (+ 320 (* (- n 6) 40)) #f `(staging right ,(- n 6)))))
    (* 2 size))
   (map-n
    (lambda (c)
      (map-n
       (lambda (r)
	 (create-square (+ 100 (* c 40))
			(+ 360 (* r 40))
			(f r c)
			`(board ,r ,c)))
       size))
    size)))


;; N&C Snap


(define snap-utterances '("play" "colour" "shape" "same" "nothing"))
(define snap-shapes '("ring" "star" "flash"))
(define snap-colours '("red" "green" "blue"))
(define snap-positions '("pos0" "pos1"))

(define (snap-decls)
  (append
   (map (lambda (u) `("utterance" (,u))) snap-utterances)
   (map (lambda (u) `("shape" (,u))) snap-shapes)
   (map (lambda (u) `("colour" (,u))) snap-colours)
   (map (lambda (u) `("position" (,u))) snap-positions)))

(define (snap1-decls)
  (append
   (list (modeh '(1 ("action" "#utterance" "+time")))
	 (modeb '(* ("state" (list (list "-shape" "-colour" "-position")
				     (list "-shape" "-colour" "-position"))
		       "+time")))
	 (modeb '(* ("state" () "+time")))
	 (modeb '(* ("successor" "+time" "-time")))
	 (positive-only-learning))
   (snap-decls)))

(define (snap2-decls)
  (append
   (list (modeh '(1 ("action" "#utterance" "+time")))
	 (modeb '(* ("state" (list (list "-shape" "-colour" "-position")) "+time")))
	 (modeb '(* ("successor" "+time" "-time")))
	 (positive-only-learning))
   (snap-decls)))

(define (snap-action c1 c2)
  (cond ((and (null? c1) (null? c2)) "")
	((and (eq? (car c1) (car c2))
	      (eq? (cadr c1) (cadr c2))) "same")
	((eq? (car c1) (car c2)) "shape")
	((eq? (cadr c1) (cadr c2)) "colour")
	(else "nothing")))

(define (generate-filled-snap1 t-prev t-curr)
  (let ((c1 `(,(random-member snap-shapes)
	      ,(random-member snap-colours)
	      ,(car snap-positions)))
	(c2 `(,(random-member snap-shapes)
	      ,(random-member snap-colours)
	      ,(cadr snap-positions))))
    (snap1->prolog `(,c1 ,c2) `(,(snap-action c1 c2)) t-curr t-prev)))

(define (generate-empty-snap1 t-prev t-curr)
  (snap1->prolog '() '("play") t-curr t-prev))

(define (number->time n) (string-append "t" (number->string n)))

(define (generate-one-snap1 n)
  (append
   (generate-empty-snap1 (number->time n) (number->time (+ n 1)))
   (generate-filled-snap1 (number->time (+ n 1)) (number->time (+ n 2)))))

(define (generate-snap1 size)
  (prolog-program->string (generate-snap1-prolog size)))

(define (generate-snap1-prolog size)
  (snap1-decls)
  (apply append
	  (map-n (lambda (x) (generate-one-snap1 (* 2 x))) size)))

(define (snap2->prolog state action time succ)
  `(("state" ((list (list ,(car state) ,(cadr state) ,(caddr state)))
	      ,time))
    ,@(if (null? action)
	  '()
	  `(("action" (,(car action) ,time))))
    ("time" (,time))
    ("successor" (,succ ,time))))

(define (snap1->prolog state action time succ)
  `(("state" (,(if (null? state)
		   `()
		   `(list (list ,(caar state) ,(cadar state) ,(caddar state))
			  (list ,(caadr state) ,(cadadr state) ,(caddadr state))))
	      ,time))
    ,@(if (null? action)
	  '()
	  `(("action" (,(car action) ,time))))
    ("time" (,time))
    ("successor" (,succ ,time))))

;; Housekeeping
