/* LaHaShem HaAretz U'Mloah */

#ifndef _TOOL1LIB_HACK_TRACK_C_H
#define _TOOL1LIB_HACK_TRACK_C_H

#include <stdarg.h>
#include <math.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sched.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <linux/videodev.h>
#include "viff.h"
#include "vdefines.h"
#include "khoros.h"
#include "pwc-ioctl.h"
#include "tool1lib-c.h"
#include "tool1lib-v4l-c.h"

#ifndef SQR
static float ___sqrarg;
#define SQR(a) (___sqrarg=(a),___sqrarg*___sqrarg)
#endif

#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef MIN
#define MIN(a,b) ((a)>(b)?(b):(a))
#endif

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define MACH_EPS 1e-7
#define MAX_FRAMES 5400
#define MAX_OBJECTS 10
#define MAX_VERTICES 300

#define IS_ZERO(x) (-MACH_EPS<(x)&&(x)<MACH_EPS)
#define DEG_TO_RAD(x) (PI/180.0*(x))

/* Global Variables */

extern int gfirst_frame[NCAMERAS], glast_frame[NCAMERAS];
extern int gobject_k[NCAMERAS][MAX_FRAMES];
extern float gellipse_x[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];
extern float gellipse_y[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];
extern float gellipse_theta[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];
extern float gellipse_r[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];
extern float gellipse_s[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];
extern int hack_track_using_v4l[NCAMERAS];
extern int hack_track_width[NCAMERAS], hack_track_height[NCAMERAS];
extern unsigned char *hack_track_buffer[NCAMERAS];
extern int hack_track_real_time;
extern int hack_track_uid;
extern int hack_track_subsample[NCAMERAS];
extern int hack_track_capturing[NCAMERAS];
extern int hack_track_tracking_ellipses[NCAMERAS];
extern int hack_track_tracking_polygons[NCAMERAS];
extern int hack_track_hull[NCAMERAS];
extern Display *hack_track_display[NCAMERAS];
extern int hack_track_screen[NCAMERAS];
extern Window hack_track_window[NCAMERAS];
extern int hack_track_window_width[NCAMERAS];
extern int  hack_track_window_height[NCAMERAS];
extern int hack_track_i[NCAMERAS], hack_track_n[NCAMERAS];
extern int  hack_track_j[NCAMERAS], hack_track_m[NCAMERAS];
extern int hack_track_k[NCAMERAS], hack_track_l[NCAMERAS];
extern GC hack_track_gc[NCAMERAS];
extern XShmSegmentInfo hack_track_shminfo[NCAMERAS];
extern XImage *hack_track_shmimage[NCAMERAS];
extern unsigned char *hack_track_frames[NCAMERAS];
extern unsigned char *hack_track_code_frames[NCAMERAS];
extern unsigned char *hack_track_last_frame[NCAMERAS];
extern unsigned char **hack_track_last_code[NCAMERAS];
extern unsigned char **hack_track_conjure_code[NCAMERAS];
extern int hack_track_starting[NCAMERAS], hack_track_conjuring[NCAMERAS];
extern int hack_track_motion_threshold[NCAMERAS];
extern int hack_track_colour_pixel_mass_threshold[NCAMERAS];
extern int hack_track_colour_proximity_threshold[NCAMERAS];
extern int hack_track_motion_pixel_mass_threshold[NCAMERAS];
extern int hack_track_motion_proximity_threshold[NCAMERAS];
extern int hack_track_rhl[NCAMERAS], hack_track_rhh[NCAMERAS];
extern int hack_track_rsl[NCAMERAS], hack_track_rsh[NCAMERAS];
extern int hack_track_rvl[NCAMERAS], hack_track_rvh[NCAMERAS];
extern int hack_track_ghl[NCAMERAS], hack_track_ghh[NCAMERAS];
extern int hack_track_gsl[NCAMERAS], hack_track_gsh[NCAMERAS];
extern int hack_track_gvl[NCAMERAS], hack_track_gvh[NCAMERAS];
extern int hack_track_bhl[NCAMERAS], hack_track_bhh[NCAMERAS];
extern int hack_track_bsl[NCAMERAS], hack_track_bsh[NCAMERAS];
extern int hack_track_bvl[NCAMERAS], hack_track_bvh[NCAMERAS];
extern int hack_track_yhl[NCAMERAS], hack_track_yhh[NCAMERAS];
extern int hack_track_ysl[NCAMERAS], hack_track_ysh[NCAMERAS];
extern int hack_track_yvl[NCAMERAS], hack_track_yvh[NCAMERAS];
extern int hack_track_first_frame[NCAMERAS];
extern int hack_track_last_moving_valid[NCAMERAS];
extern int **hack_track_last_moving[NCAMERAS];
extern int hack_track_interrupts_on;
extern unsigned short hack_track_polygon_l[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];
extern unsigned short hack_track_polygon_x
[NCAMERAS][MAX_FRAMES][MAX_OBJECTS][MAX_VERTICES];
extern unsigned short hack_track_polygon_y
[NCAMERAS][MAX_FRAMES][MAX_OBJECTS][MAX_VERTICES];
extern int hack_track_motion[NCAMERAS][MAX_FRAMES];
extern unsigned char hack_track_object_colour
[NCAMERAS][MAX_FRAMES][MAX_OBJECTS];

/* Function Prototypes */

void panic(const char *error_text, ...);
int get_first_frame(int camera);
int get_last_frame(int camera);
int get_object_k(int camera, int frame);
float get_ellipse_x(int camera, int frame, int k);
float get_ellipse_y(int camera, int frame, int k);
float get_ellipse_theta(int camera, int frame, int k);
float get_ellipse_r(int camera, int frame, int k);
float get_ellipse_s(int camera, int frame, int k);
float my_atan2(float x, float y);
float quadratic1(float a, float b, float c);
float quadratic2(float a, float b, float c);
float eigenvalue1(float xx, float xy, float yx, float yy);
float eigenvalue2(float xx, float xy, float yx, float yy);
float eigenvector_angle1(float xx, float xy, float yx, float yy);
float eigenvector_angle2(float xx, float xy, float yx, float yy);
int hack_track_get_using_v4l(int camera);
int hack_track_get_current_frame(int camera);
int hack_track_get_motion(int camera, int frame);
int hack_track_get_object_colour(int camera, int frame, int k);
unsigned hack_track_get_polygon_l(int camera, int frame, int k);
unsigned hack_track_get_polygon_x(int camera, int frame, int k, int l);
unsigned hack_track_get_polygon_y(int camera, int frame, int k, int l);
int hack_track_angle_between(int l, int a, int h);
void hack_track_tracker(int camera,
			unsigned char *buffer,
			int frame,
			int height,
			int width,
			int capturing,
			int capturing_code,
			int tracking_ellipses,
			int tracking_polygons,
			int hull);
int hack_track_interrupts(void);
void hack_track_set_interrupts(int on);
void hack_track_handler(int signum);
void hack_track_offline_tracker(int camera,
				int frame,
				int height,
				int width,
				int subsample,
				int capturing_code,
				int tracking_ellipses,
				int tracking_polygons,
				int hull);
void hack_track_draw_ellipse(int camera,
			     float x0,
			     float y0,
			     float t0,
			     float a,
			     float b);
void hack_track_set_real_time(void);
void hack_track_set_nonreal_time(void);
int hack_track_get_red(int camera, int i, int y, int x);
int hack_track_get_green(int camera, int i, int y, int x);
int hack_track_get_blue(int camera, int i, int y, int x);
int hack_track_get_pixel(int camera, int i, int y, int x);
void hack_track_set_red(int camera, int i, int y, int x, int v);
void hack_track_set_green(int camera, int i, int y, int x, int v);
void hack_track_set_blue(int camera, int i, int y, int x, int v);
void hack_track_set_pixel(int camera, int i, int y, int x, int v);
int hack_track_get_code_frame_red(int camera, int i, int y, int x);
int hack_track_get_code_frame_green(int camera, int i, int y, int x);
int hack_track_get_code_frame_blue(int camera, int i, int y, int x);
void hack_track_copy_frame_to_ximage(int camera, int i, XImage *ximage);
void hack_track_motion_detector(int camera, int i, XImage *ximage);
void hack_track_set_thresholds(int camera,
			       int motion_threshold,
			       int colour_pixel_mass_threshold,
			       int colour_proximity_threshold,
			       int motion_pixel_mass_threshold,
			       int motion_proximity_threshold);
void hack_track_set_red_range(int camera,
			      int hl,
			      int hh,
			      int sl,
			      int sh,
			      int vl,
			      int vh);
void hack_track_set_green_range(int camera,
				int hl,
				int hh,
				int sl,
				int sh,
				int vl,
				int vh);
void hack_track_set_blue_range(int camera,
			       int hl,
			       int hh,
			       int sl,
			       int sh,
			       int vl,
			       int vh);
void hack_track_set_yellow_range(int camera,
				 int hl,
				 int hh,
				 int sl,
				 int sh,
				 int vl,
				 int vh);
void hack_track_start(int camera, int frame);
void hack_track_step(int camera, int frame, int object_frame);
void hack_track_finish(int camera);
void hack_track_show_frame(int camera, int i);
void hack_track_show_code_frame(int camera, int i);
void hack_track_free(int camera);
void hack_track_attach_viewfinder(int camera,
				  Display *display,
				  int screen,
				  Window window,
				  int window_x,
				  int window_y,
				  int window_width,
				  int window_height);
void hack_track_detach_viewfinder(int camera);
void hack_track_sigaction(int camera);
void hack_track_block(void);
void hack_track_unblock(void);
void hack_track_turn_on_viewfinder(int camera, int m);
void hack_track_turn_on_capturing(int camera, int i, int l, int n);
void hack_track_turn_on_ellipse_tracker(int camera, int i, int l, int n);
void hack_track_turn_on_ellipse_tracker_with_capturing(int camera,
						       int i,
						       int l,
						       int n);
void hack_track_turn_on_polygon_tracker(int camera,
					int i,
					int l,
					int n,
					int h);
void hack_track_turn_on_polygon_tracker_with_capturing(int camera,
						       int i,
						       int l,
						       int n,
						       int h);
void hack_track_turn_off(int camera);
void hack_track_capture_frames(int camera, int i, int l, int n);
void hack_track_track_ellipses(int camera, int i, int l, int n);
void hack_track_track_ellipses_and_capture_frames(int camera,
						  int i,
						  int l,
						  int n);
void hack_track_track_polygons(int camera, int i, int l, int n, int h);
void hack_track_track_polygons_and_capture_frames(int camera,
						  int i,
						  int l,
						  int n,
						  int h);
void hack_track_malloc(int camera, int nframes);
void hack_track_startup(int camera,
			int height,
			int width,
			int nframes,
			int subsample,
			int raw_frame_rate);
void hack_track_startup_offline(int camera,
				int height,
				int width,
				int nframes,
				int subsample);
void hack_track_startup_with_viewfinder(int camera,
					int height,
					int width,
					int nframes,
					int subsample,
					Display *display,
					int screen,
					Window window,
					int window_x,
					int window_y,
					int window_width,
					int window_height,
					int raw_frame_rate);
void hack_track_restartup(int camera,
			  int height,
			  int width,
			  int nframes,
			  int subsample,
			  int raw_frame_rate);
void hack_track_page_in_frames(int camera, int iframes, int nframes);
void hack_track_shutdown(int camera);
void hack_track_shutdown_offline(int camera);
void hack_track_send_frame(int camera, FILE *f, int i);
void hack_track_receive_frame(int camera, FILE *f, int i);
void hack_track_handler_multiple(int signum);
void hack_track_sigaction_multiple(int ncameras);
void hack_track_startup_multiple(int ncameras,
				 int height,
				 int width,
				 int nframes,
				 int subsample,
				 int raw_frame_rate);
void hack_track_startup_with_viewfinder_multiple(int ncameras,
						 int height,
						 int width,
						 int nframes,
						 int subsample,
						 Display *display,
						 int screen,
						 Window window,
						 int window_x[NCAMERAS],
						 int window_y[NCAMERAS],
						 int window_width,
						 int window_height,
						 int raw_frame_rate);
void hack_track_restartup_multiple(int ncameras,
				   int height,
				   int width,
				   int nframes,
				   int subsample,
				   int raw_frame_rate);
void hack_track_shutdown_multiple(int ncameras);
#endif

/* Tam V'Nishlam Shevah L'El Borei Olam */
