/* LaHaShem HaAretz U'Mloah */
/* $Id: tool1lib-v4l-c.c,v 1.3 2008-11-13 18:42:47 qobi Exp $ */

/* no VIDIOCSTUNER */
/* no VIDIOCGCHAN */
/* no VIDIOCSCHAN */

/* Includes */

#include "tool1lib-v4l-c.h"

/* Global Variables */

char v4l_video_device[NCAMERAS][100] = {"/dev/video0",
					"/dev/video1",
					"/dev/video2",
					"/dev/video3"};
int v4l_fd[NCAMERAS];
int v4l_width[NCAMERAS], v4l_height[NCAMERAS], v4l_palette[NCAMERAS];
unsigned char *v4l_frame[NCAMERAS], *v4l_converted_frame[NCAMERAS];

/* Until ccvt is ported to x86_64 fall back to a C version */

#ifndef __i386__

int clamp(int v, int l, int u)
{
  return ((v < l) ? l : (v > u ? u : v));
}

void yuvToBGRpixel(int y,
		   int u,
		   int v,
		   unsigned char* b,
		   unsigned char* g,
		   unsigned char* r)
{
  int r_ = y + 1.402 * (v-128);
  int g_ = y - 0.34414 * (u-128) - 0.71414 * (v-128);
  int b_ = y + 1.772 * (u-128);

  *r = clamp(r_, 0, 255);
  *g = clamp(g_, 0, 255);
  *b = clamp(b_, 0, 255);
}

void ccvt_420p_bgr32(const int width, const int height,
		     const unsigned char *srcy, const unsigned char *srcu,
		     const unsigned char *srcv, unsigned char *dst)
{
  int i;
  const size = width*height;

  for(i = 0; i < size; ++i)
  {
    yuvToBGRpixel(srcy[i],
		  srcu[(width/2)*(i/width/2) + (i%width)/2],
		  srcv[(width/2)*(i/width/2) + (i%width)/2],
		  &dst[i*4],
		  &dst[i*4+1],
		  &dst[i*4+2]);
    dst[i*4+3] = 0;
  }
}

#endif

/* Function Definitions */

void v4l_set_video_device(int camera, char *device)
{ if (strlen(device)>=99)
  { perror("v4l set_video_device device name too long");
    exit(-1);}
  strncpy(&v4l_video_device[camera][0], device, 100);}

int v4l_available(int camera)
{ int fd;
  fd = open(&v4l_video_device[camera][0], O_RDONLY);
  if (fd<0) return FALSE;
  close(fd);
  return TRUE;}

void v4l_open_video(int camera)
{ v4l_fd[camera] = open(&v4l_video_device[camera][0], O_RDONLY);
  if (v4l_fd[camera]<0)
  { perror("v4l open failed");
    exit(-1);}
  v4l_width[camera] = v4l_get_width(camera);
  v4l_height[camera] = v4l_get_height(camera);
  v4l_palette[camera] = v4l_get_palette(camera);}

void v4l_close_video(int camera)
{ if (close(v4l_fd[camera])!=0)
  { perror("v4l close failed");
    exit(-1);}}

void v4l_restore_factory_settings(int camera)
{ if (ioctl(v4l_fd[camera], VIDIOCPWCFACTORY)!=0)
  { perror("v4l ioctl VIDIOCPWCFACTORY failed");
    exit(-1);}}

void v4l_save_user_settings(int camera)
{ if (ioctl(v4l_fd[camera], VIDIOCPWCSUSER)!=0)
  { perror("v4l ioctl VIDIOCPWCSUSER failed");
    exit(-1);}}

void v4l_restore_user_settings(int camera)
{ if (ioctl(v4l_fd[camera], VIDIOCPWCRUSER)!=0)
  { perror("v4l ioctl VIDIOCPWCRUSER failed");
    exit(-1);}}

int v4l_get_agc(int camera)
{ int agc;
  if (ioctl(v4l_fd[camera], VIDIOCPWCGAGC, &agc)!=0)
  { perror("v4l ioctl VIDIOCPWCGAGC failed");
    exit(-1);}
  return agc;}

void v4l_set_agc(int camera, int agc)
{ if (ioctl(v4l_fd[camera], VIDIOCPWCSAGC, &agc)!=0)
  { perror("v4l ioctl VIDIOCPWCSAGC failed");
    exit(-1);}}

void v4l_set_shutter(int camera, int shutter)
{ if (ioctl(v4l_fd[camera], VIDIOCPWCSSHUTTER, &shutter)!=0)
  { perror("v4l ioctl VIDIOCPWCSSHUTTER failed");
    exit(-1);}}

int v4l_get_compression(int camera)
{ int compression;
  if (ioctl(v4l_fd[camera], VIDIOCPWCGCQUAL, &compression)!=0)
  { perror("v4l ioctl VIDIOCPWCGCQUAL failed");
    exit(-1);}
  return compression;}

void v4l_set_compression(int camera, int compression)
{ if (ioctl(v4l_fd[camera], VIDIOCPWCSCQUAL, &compression)!=0)
  { perror("v4l ioctl VIDIOCPWCSCQUAL failed");
    exit(-1);}}

int v4l_get_palette(int camera)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  return video_picture.palette;}

void v4l_set_palette(int camera, int palette, int depth)
{ struct video_picture video_picture;
  v4l_palette[camera] = palette;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  video_picture.palette = palette;
  video_picture.depth = depth;
  if (ioctl(v4l_fd[camera], VIDIOCSPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCSPICT failed");
    exit(-1);}}

int v4l_get_brightness(int camera)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  return video_picture.brightness;}

void v4l_set_brightness(int camera, int brightness)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  video_picture.brightness = brightness&0xffff;
  if (ioctl(v4l_fd[camera], VIDIOCSPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCSPICT failed");
    exit(-1);}}

int v4l_get_contrast(int camera)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  return video_picture.contrast;}

void v4l_set_contrast(int camera, int contrast)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  video_picture.contrast = contrast&0xffff;
  if (ioctl(v4l_fd[camera], VIDIOCSPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCSPICT failed");
    exit(-1);}}

int v4l_get_hue(int camera)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  return video_picture.hue;}

void v4l_set_hue(int camera, int hue)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  video_picture.hue = hue&0xffff;
  if (ioctl(v4l_fd[camera], VIDIOCSPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCSPICT failed");
    exit(-1);}}

int v4l_get_colour(int camera)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  return video_picture.colour;}

void v4l_set_colour(int camera, int colour)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  video_picture.colour = colour&0xffff;
  if (ioctl(v4l_fd[camera], VIDIOCSPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCSPICT failed");
    exit(-1);}}

int v4l_get_whiteness(int camera)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  return video_picture.whiteness;}

void v4l_set_whiteness(int camera, int whiteness)
{ struct video_picture video_picture;
  if (ioctl(v4l_fd[camera], VIDIOCGPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCGPICT failed");
    exit(-1);}
  video_picture.whiteness = whiteness&0xffff;
  if (ioctl(v4l_fd[camera], VIDIOCSPICT, &video_picture)!=0)
  { perror("v4l ioctl VIDIOCSPICT failed");
    exit(-1);}}

int v4l_get_width(int camera)
{ struct video_window video_window;
  memset(&video_window, 0, sizeof(struct video_window));
  if (ioctl(v4l_fd[camera], VIDIOCGWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCGWIN failed");
    exit(-1);}
  return video_window.width;}

int v4l_get_height(int camera)
{ struct video_window video_window;
  memset(&video_window, 0, sizeof(struct video_window));
  if (ioctl(v4l_fd[camera], VIDIOCGWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCGWIN failed");
    exit(-1);}
  return video_window.height;}

void v4l_set_size(int camera, int width, int height)
{ struct video_window video_window;
  v4l_width[camera] = width;
  v4l_height[camera] = height;
  memset(&video_window, 0, sizeof(struct video_window));
  if (ioctl(v4l_fd[camera], VIDIOCGWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCGWIN failed");
    exit(-1);}
  video_window.width = width;
  video_window.height = height;
  video_window.clipcount = 0;
  if (ioctl(v4l_fd[camera], VIDIOCSWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCSWIN failed");
    exit(-1);}}

int v4l_get_frame_rate(int camera)
{ struct video_window video_window;
  memset(&video_window, 0, sizeof(struct video_window));
  if (ioctl(v4l_fd[camera], VIDIOCGWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCGWIN failed");
    exit(-1);}
  return (video_window.flags&PWC_FPS_FRMASK)>>PWC_FPS_SHIFT;}

void v4l_set_frame_rate(int camera, int frame_rate, int snap)
{ struct video_window video_window;
  memset(&video_window, 0, sizeof(struct video_window));
  if (ioctl(v4l_fd[camera], VIDIOCGWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCGWIN failed");
    exit(-1);}
  video_window.flags =
  (video_window.flags&~PWC_FPS_MASK)|(frame_rate<<PWC_FPS_SHIFT);
  if (snap) video_window.flags|=PWC_FPS_SNAPSHOT;
  if (ioctl(v4l_fd[camera], VIDIOCSWIN, &video_window)!=0)
  { perror("v4l ioctl VIDIOCSWIN failed");
    exit(-1);}}

int v4l_get_maxwidth(int camera)
{ struct video_capability video_capability;
  if (ioctl(v4l_fd[camera], VIDIOCGCAP, &video_capability)!=0)
  { perror("v4l ioctl VIDIOCGCAP failed");
    exit(-1);}
  return video_capability.maxwidth;}

int v4l_get_maxheight(int camera)
{ struct video_capability video_capability;
  if (ioctl(v4l_fd[camera], VIDIOCGCAP, &video_capability)!=0)
  { perror("v4l ioctl VIDIOCGCAP failed");
    exit(-1);}
  return video_capability.maxheight;}

int v4l_get_minwidth(int camera)
{ struct video_capability video_capability;
  if (ioctl(v4l_fd[camera], VIDIOCGCAP, &video_capability)!=0)
  { perror("v4l ioctl VIDIOCGCAP failed");
    exit(-1);}
  return video_capability.minwidth;}

int v4l_get_minheight(int camera)
{ struct video_capability video_capability;
  if (ioctl(v4l_fd[camera], VIDIOCGCAP, &video_capability)!=0)
  { perror("v4l ioctl VIDIOCGCAP failed");
    exit(-1);}
  return video_capability.minheight;}

int v4l_get_channels(int camera)
{ struct video_capability video_capability;
  if (ioctl(v4l_fd[camera], VIDIOCGCAP, &video_capability)!=0)
  { perror("v4l ioctl VIDIOCGCAP failed");
    exit(-1);}
  return video_capability.channels;}

char *v4l_get_name(int camera)
{ struct video_capability video_capability;
  static char name[33];
  if (ioctl(v4l_fd[camera], VIDIOCGCAP, &video_capability)!=0)
  { perror("v4l ioctl VIDIOCGCAP failed");
    exit(-1);}
  strncpy(name, video_capability.name, 32);
  name[32] = '\0';
  return &name[0];}

void v4l_allocate_frame(int camera)
{ v4l_frame[camera] =
  (unsigned char *)malloc(v4l_width[camera]*v4l_height[camera]*4);
  if (v4l_frame[camera]==NULL)
  { fprintf(stderr, "v4l malloc failed");
    exit(-1);}
  v4l_converted_frame[camera] =
  (unsigned char *)malloc(v4l_width[camera]*v4l_height[camera]*4);
  if (v4l_converted_frame[camera]==NULL)
  { fprintf(stderr, "v4l malloc failed");
    exit(-1);}}

void v4l_free_frame(int camera)
{ free(v4l_frame[camera]);
  free(v4l_converted_frame[camera]);}

int v4l_frame_available(int camera)
{ fd_set readfds;
  struct timeval timeval;
  int retval;
  FD_ZERO(&readfds);
  FD_SET(v4l_fd[camera], &readfds);
  timeval.tv_sec = 0;
  timeval.tv_usec = 0;
  retval = select(v4l_fd[camera]+1, &readfds, NULL, NULL, &timeval);
  if (retval==-1)
  { perror("v4l select failed");
    exit(-1);}
  return retval;}

void v4l_get_frame(int camera)
{ unsigned char *dy, *du, *dv;
  int n;
  if (read(v4l_fd[camera],
           v4l_frame[camera],
           v4l_width[camera]*v4l_height[camera]*4)<0)
  { perror("v4l read failed");
    exit(-1);}
  n = v4l_width[camera]*v4l_height[camera];
  dy = v4l_frame[camera];
  du = v4l_frame[camera]+n;
  dv = du+(n/4);
  ccvt_420p_bgr32(v4l_width[camera],
		  v4l_height[camera],
		  dy,
		  du,
		  dv,
		  v4l_converted_frame[camera]);}

void v4l_mmap_frame(int camera)
{ struct video_mbuf video_mbuf;
  if (ioctl(v4l_fd[camera], VIDIOCGMBUF, &video_mbuf)!=0)
  { perror("v4l ioctl VIDIOCGMBUF failed");
    exit(-1);}
  v4l_frame[camera] = (unsigned char *) mmap(NULL,
					     video_mbuf.size,
					     PROT_READ,
					     MAP_SHARED,
					     v4l_fd[camera],
					     0);
  if (v4l_frame[camera]==(unsigned char *)-1)
  { perror("v4l mmap failed");
    exit(-1);}
  v4l_converted_frame[camera] =
  (unsigned char *)malloc(v4l_width[camera]*v4l_height[camera]*4);
  if (v4l_converted_frame[camera]==NULL)
  { fprintf(stderr, "v4l malloc failed");
    exit(-1);}}

void v4l_munmap_frame(int camera)
{ struct video_mbuf video_mbuf;
  if (ioctl(v4l_fd[camera], VIDIOCGMBUF, &video_mbuf)!=0)
  { perror("v4l ioctl VIDIOCGMBUF failed");
    exit(-1);}
  munmap(v4l_frame[camera], video_mbuf.size);
  free(v4l_converted_frame[camera]);}

int v4l_get_number_of_mmaped_frames(int camera)
{ struct video_mbuf video_mbuf;
  if (ioctl(v4l_fd[camera], VIDIOCGMBUF, &video_mbuf)!=0)
  { perror("v4l ioctl VIDIOCGMBUF failed");
    exit(-1);}
  return video_mbuf.frames;}

void v4l_mmap_capture(int camera, int frame)
{ struct video_mmap video_mmap;
  video_mmap.frame = frame;
  video_mmap.format = v4l_palette[camera];
  video_mmap.width = v4l_width[camera];
  video_mmap.height = v4l_height[camera];
  if (ioctl(v4l_fd[camera], VIDIOCMCAPTURE, &video_mmap)<0)
  { perror("v4l ioctl VIDIOCMCAPTURE failed");
    exit(-1);}}

void v4l_mmap_sync(int camera, int frame)
{ if (ioctl(v4l_fd[camera], VIDIOCSYNC, &frame)<0)
  { perror("v4l ioctl VIDIOCSYNC failed");
    exit(-1);}}

void v4l_mmap_get_frame(int camera)
{ unsigned char *dy, *du, *dv;
  int n;
  /* needs work: to handle video_mbuf.offsets[frame] */
  n = v4l_width[camera]*v4l_height[camera];
  dy = v4l_frame[camera];
  du = v4l_frame[camera]+n;
  dv = du+(n/4);
  ccvt_420p_bgr32(v4l_width[camera],
		  v4l_height[camera],
		  dy,
		  du,
		  dv,
		  v4l_converted_frame[camera]);}

void v4l_interrupt_on(int frame_rate)
{ struct itimerval itimerval;
  itimerval.it_interval.tv_sec = 0;
  itimerval.it_interval.tv_usec = 1000000/frame_rate;
  itimerval.it_value.tv_sec = 0;
  itimerval.it_value.tv_usec = 1000000/frame_rate;
  setitimer(ITIMER_REAL, &itimerval, (struct itimerval *)0);}

void v4l_interrupt_off(void)
{ struct itimerval itimerval;
  itimerval.it_interval.tv_sec = 0;
  itimerval.it_interval.tv_usec = 0;
  itimerval.it_value.tv_sec = 0;
  itimerval.it_value.tv_usec = 0;
  setitimer(ITIMER_REAL, &itimerval, (struct itimerval *)0);}

void v4l_write_ppm(int camera, char *name)
{ FILE *f;
  int y, x;
  f = fopen(name, "w");
  fprintf(f, "P6\n");
  fprintf(f, "%d %d\n", v4l_width[camera], v4l_height[camera]);
  fprintf(f, "255\n");
  for (y = 0; y<v4l_height[camera]; y++)
  { for (x = 0; x<v4l_width[camera]; x++)
    { putc(v4l_converted_frame[camera][(y*v4l_width[camera]+x)*4+2], f);
      putc(v4l_converted_frame[camera][(y*v4l_width[camera]+x)*4+1], f);
      putc(v4l_converted_frame[camera][(y*v4l_width[camera]+x)*4+0], f);}}
  fclose(f);}

void v4l_home(int camera, int tilt, int pan)
{ int function = tilt<<1|pan;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTRESET, &function)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTRESET failed");
    exit(-1);}
  return;}

int v4l_get_pan_min(int camera)
{ struct pwc_mpt_range pwc_mpt_range;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTGRANGE, &pwc_mpt_range)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTGRANGE failed");
    exit(-1);}
  return pwc_mpt_range.pan_min;}

int v4l_get_pan_max(int camera)
{ struct pwc_mpt_range pwc_mpt_range;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTGRANGE, &pwc_mpt_range)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTGRANGE failed");
    exit(-1);}
  return pwc_mpt_range.pan_max;}

int v4l_get_tilt_min(int camera)
{ struct pwc_mpt_range pwc_mpt_range;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTGRANGE, &pwc_mpt_range)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTGRANGE failed");
    exit(-1);}
  return pwc_mpt_range.tilt_min;}

int v4l_get_tilt_max(int camera)
{ struct pwc_mpt_range pwc_mpt_range;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTGRANGE, &pwc_mpt_range)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTGRANGE failed");
    exit(-1);}
  return pwc_mpt_range.tilt_max;}

int v4l_get_pan(int camera)
{ struct pwc_mpt_angles pwc_mpt_angles;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTGANGLE, &pwc_mpt_angles)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTGANGLE failed");
    exit(-1);}
  return pwc_mpt_angles.pan;}

int v4l_get_tilt(int camera)
{ struct pwc_mpt_angles pwc_mpt_angles;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTGANGLE, &pwc_mpt_angles)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTGANGLE failed");
    exit(-1);}
  return pwc_mpt_angles.tilt;}

void v4l_set_pan_tilt(int camera, int pan, int tilt, int absolute)
{ struct pwc_mpt_angles pwc_mpt_angles;
  pwc_mpt_angles.absolute = absolute;
  pwc_mpt_angles.pan = pan;
  pwc_mpt_angles.tilt = tilt;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTSANGLE, &pwc_mpt_angles)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTSANGLE failed");
    exit(-1);}
  return;}

int v4l_pan_tilt_status(int camera)
{ struct pwc_mpt_status pwc_mpt_status;
  if (ioctl(v4l_fd[camera], VIDIOCPWCMPTSTATUS, &pwc_mpt_status)!=0)
  { perror("v4l ioctl VIDIOCPWCMPTSTATUS failed");
    exit(-1);}
  return pwc_mpt_status.status;}

/* Tam V'Nishlam Shevah L'El Borei Olam */
